# -*- encoding: utf-8 -*-
# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from http import client as http_client
import io
from unittest import mock

from oslo_config import cfg
from oslo_utils import uuidutils

from ironic import api
from ironic.api.controllers.v1 import node as api_node
from ironic.api.controllers.v1 import utils
from ironic.common import context as ironic_context
from ironic.common import exception
from ironic.common import policy
from ironic.common import states
from ironic import objects
from ironic.tests import base
from ironic.tests.unit.api import utils as test_api_utils
from ironic.tests.unit.objects import utils as obj_utils

CONF = cfg.CONF


class TestApiUtils(base.TestCase):

    def test_validate_limit(self):
        limit = utils.validate_limit(10)
        self.assertEqual(10, limit)

        # max limit
        limit = utils.validate_limit(999999999)
        self.assertEqual(CONF.api.max_limit, limit)

        # negative
        self.assertRaises(exception.ClientSideError, utils.validate_limit, -1)

        # zero
        self.assertRaises(exception.ClientSideError, utils.validate_limit, 0)

    def test_validate_sort_dir(self):
        sort_dir = utils.validate_sort_dir('asc')
        self.assertEqual('asc', sort_dir)

        # invalid sort_dir parameter
        self.assertRaises(exception.ClientSideError,
                          utils.validate_sort_dir,
                          'fake-sort')

    def test_apply_jsonpatch(self):
        doc = {"foo": {"bar": "baz"}}
        patch = [{"op": "add", "path": "/foo/answer", "value": 42}]
        result = utils.apply_jsonpatch(doc, patch)
        expected = {"foo": {"bar": "baz", "answer": 42}}
        self.assertEqual(expected, result)

    def test_apply_jsonpatch_no_add_root_attr(self):
        doc = {}
        patch = [{"op": "add", "path": "/foo", "value": 42}]
        self.assertRaisesRegex(exception.ClientSideError,
                               "Adding a new attribute",
                               utils.apply_jsonpatch, doc, patch)

    def test_apply_jsonpatch_remove_non_existent(self):
        # Raises a KeyError.
        doc = {}
        patch = [{"op": "remove", "path": "/foo"}]
        self.assertRaises(exception.PatchError,
                          utils.apply_jsonpatch, doc, patch)

    def test_apply_jsonpatch_replace_non_existent_list_item(self):
        # Raises an IndexError.
        doc = []
        patch = [{"op": "replace", "path": "/0", "value": 42}]
        self.assertRaises(exception.PatchError,
                          utils.apply_jsonpatch, doc, patch)

    def test_get_patch_values_no_path(self):
        patch = [{'path': '/name', 'op': 'update', 'value': 'node-0'}]
        path = '/invalid'
        values = utils.get_patch_values(patch, path)
        self.assertEqual([], values)

    def test_get_patch_values_remove(self):
        patch = [{'path': '/name', 'op': 'remove'}]
        path = '/name'
        values = utils.get_patch_values(patch, path)
        self.assertEqual([], values)

    def test_get_patch_values_success(self):
        patch = [{'path': '/name', 'op': 'replace', 'value': 'node-x'}]
        path = '/name'
        values = utils.get_patch_values(patch, path)
        self.assertEqual(['node-x'], values)

    def test_get_patch_values_multiple_success(self):
        patch = [{'path': '/name', 'op': 'replace', 'value': 'node-x'},
                 {'path': '/name', 'op': 'replace', 'value': 'node-y'}]
        path = '/name'
        values = utils.get_patch_values(patch, path)
        self.assertEqual(['node-x', 'node-y'], values)

    def test_is_path_removed_success(self):
        patch = [{'path': '/name', 'op': 'remove'}]
        path = '/name'
        value = utils.is_path_removed(patch, path)
        self.assertTrue(value)

    def test_is_path_removed_subpath_success(self):
        patch = [{'path': '/local_link_connection/switch_id', 'op': 'remove'}]
        path = '/local_link_connection'
        value = utils.is_path_removed(patch, path)
        self.assertTrue(value)

    def test_is_path_removed_similar_subpath(self):
        patch = [{'path': '/local_link_connection_info/switch_id',
                  'op': 'remove'}]
        path = '/local_link_connection'
        value = utils.is_path_removed(patch, path)
        self.assertFalse(value)

    def test_is_path_removed_replace(self):
        patch = [{'path': '/name', 'op': 'replace', 'value': 'node-x'}]
        path = '/name'
        value = utils.is_path_removed(patch, path)
        self.assertFalse(value)

    def test_is_path_updated_success(self):
        patch = [{'path': '/name', 'op': 'remove'}]
        path = '/name'
        value = utils.is_path_updated(patch, path)
        self.assertTrue(value)

    def test_is_path_updated_subpath_success(self):
        patch = [{'path': '/properties/switch_id', 'op': 'add', 'value': 'id'}]
        path = '/properties'
        value = utils.is_path_updated(patch, path)
        self.assertTrue(value)

    def test_is_path_updated_similar_subpath(self):
        patch = [{'path': '/properties2/switch_id',
                  'op': 'replace', 'value': 'spam'}]
        path = '/properties'
        value = utils.is_path_updated(patch, path)
        self.assertFalse(value)

    def test_check_for_invalid_fields(self):
        requested = ['field_1', 'field_3']
        supported = ['field_1', 'field_2', 'field_3']
        utils.check_for_invalid_fields(requested, supported)

    def test_check_for_invalid_fields_fail(self):
        requested = ['field_1', 'field_4']
        supported = ['field_1', 'field_2', 'field_3']
        self.assertRaises(exception.InvalidParameterValue,
                          utils.check_for_invalid_fields,
                          requested, supported)

    def test_patch_update_changed_fields(self):
        schema = {
            'properties': {
                'one': {},
                'two': {},
                'three': {},
                'four': {},
                'five_uuid': {},
            }
        }
        fields = [
            'one',
            'two',
            'three',
            'four',
            'five_id'
        ]

        def rpc_object():
            obj = mock.MagicMock()
            items = {
                'one': 1,
                'two': 'ii',
                'three': None,
                'four': [1, 2, 3, 4],
                'five_id': 123
            }
            obj.__getitem__.side_effect = items.__getitem__
            obj.__contains__.side_effect = items.__contains__
            return obj

        # test no change
        o = rpc_object()
        utils.patch_update_changed_fields({
            'one': 1,
            'two': 'ii',
            'three': None,
            'four': [1, 2, 3, 4],
        }, o, fields, schema, id_map={'five_id': 123})
        o.__setitem__.assert_not_called()

        # test everything changes, and id_map values override from_dict values
        o = rpc_object()
        utils.patch_update_changed_fields({
            'one': 2,
            'two': 'iii',
            'three': '',
            'four': [2, 3],
        }, o, fields, schema, id_map={'four': [4], 'five_id': 456})
        o.__setitem__.assert_has_calls([
            mock.call('one', 2),
            mock.call('two', 'iii'),
            mock.call('three', ''),
            mock.call('four', [4]),
            mock.call('five_id', 456)
        ])

        # test None fields from None values and missing keys
        # also five_id is untouched with no id_map
        o = rpc_object()
        utils.patch_update_changed_fields({
            'two': None,
        }, o, fields, schema)
        o.__setitem__.assert_has_calls([
            mock.call('two', None),
        ])

        # test fields not in the schema are untouched
        fields = [
            'six',
            'seven',
            'eight'
        ]
        o = rpc_object()
        utils.patch_update_changed_fields({
            'six': 2,
            'seven': 'iii',
            'eight': '',
        }, o, fields, schema)
        o.__setitem__.assert_not_called()

    def test_patched_validate_with_schema(self):
        schema = {
            'properties': {
                'one': {'type': 'string'},
                'two': {'type': 'integer'},
                'three': {'type': 'boolean'},
            }
        }

        # test non-schema fields removed
        pd = {
            'one': 'one',
            'two': 2,
            'three': True,
            'four': 4,
            'five': 'five'
        }
        utils.patched_validate_with_schema(pd, schema)
        self.assertEqual({
            'one': 'one',
            'two': 2,
            'three': True,
        }, pd)

        # test fails schema validation
        pd = {
            'one': 1,
            'two': 2,
            'three': False
        }
        e = self.assertRaises(exception.InvalidParameterValue,
                              utils.patched_validate_with_schema, pd, schema)
        self.assertIn("1 is not of type 'string'", str(e))

        # test fails custom validation
        def validate(name, value):
            raise exception.InvalidParameterValue('big ouch')

        pd = {
            'one': 'one',
            'two': 2,
            'three': False
        }
        e = self.assertRaises(exception.InvalidParameterValue,
                              utils.patched_validate_with_schema, pd, schema,
                              validate)
        self.assertIn("big ouch", str(e))

    def test_patch_validate_allowed_fields(self):
        allowed_fields = ['one', 'two', 'three']

        # patch all
        self.assertEqual(
            {'one', 'two', 'three'},
            utils.patch_validate_allowed_fields([
                {'path': '/one'},
                {'path': '/two'},
                {'path': '/three/four'},
            ], allowed_fields))

        # patch one
        self.assertEqual(
            {'one'},
            utils.patch_validate_allowed_fields([
                {'path': '/one'},
            ], allowed_fields))

        # patch invalid field
        e = self.assertRaises(
            exception.Invalid,
            utils.patch_validate_allowed_fields,
            [{'path': '/four'}],
            allowed_fields)
        self.assertIn("Cannot patch /four. "
                      "Only the following can be updated: "
                      "one, two, three", str(e))

    @mock.patch.object(api, 'request', autospec=False)
    def test_sanitize_dict(self, mock_req):
        mock_req.public_url = 'http://192.0.2.1:5050'

        node = obj_utils.get_test_node(
            self.context,
            created_at=datetime.datetime(2000, 1, 1, 0, 0),
            updated_at=datetime.datetime(2001, 1, 1, 0, 0),
            inspection_started_at=datetime.datetime(2002, 1, 1, 0, 0),
            console_enabled=True,
            tags=['one', 'two', 'three'])

        expected_links = [{
            'href': 'http://192.0.2.1:5050/v1/node/'
                    '1be26c0b-03f2-4d2e-ae87-c02d7f33c123',
            'rel': 'self'
        }, {
            'href': 'http://192.0.2.1:5050/node/'
                    '1be26c0b-03f2-4d2e-ae87-c02d7f33c123',
            'rel': 'bookmark'
        }]

        # all fields
        node_dict = utils.object_to_dict(
            node,
            link_resource='node',
        )
        utils.sanitize_dict(node_dict, None)
        self.assertEqual({
            'created_at': '2000-01-01T00:00:00+00:00',
            'links': expected_links,
            'updated_at': '2001-01-01T00:00:00+00:00',
            'uuid': '1be26c0b-03f2-4d2e-ae87-c02d7f33c123'
        }, node_dict)

        # some fields
        node_dict = utils.object_to_dict(
            node,
            link_resource='node',
        )
        utils.sanitize_dict(node_dict, ['uuid', 'created_at'])
        self.assertEqual({
            'created_at': '2000-01-01T00:00:00+00:00',
            'links': expected_links,
            'uuid': '1be26c0b-03f2-4d2e-ae87-c02d7f33c123'
        }, node_dict)

        # no fields
        node_dict = utils.object_to_dict(
            node,
            link_resource='node',
        )
        utils.sanitize_dict(node_dict, [])
        self.assertEqual({
            'links': expected_links,
        }, node_dict)


@mock.patch.object(api, 'request', spec_set=['version'])
class TestCheckAllowFields(base.TestCase):

    def test_check_allow_specify_fields(self, mock_request):
        mock_request.version.minor = 8
        self.assertIsNone(utils.check_allow_specify_fields(['foo']))

    def test_check_allow_specify_fields_fail(self, mock_request):
        mock_request.version.minor = 7
        self.assertRaises(exception.NotAcceptable,
                          utils.check_allow_specify_fields, ['foo'])

    def test_check_allowed_fields_network_interface(self, mock_request):
        mock_request.version.minor = 20
        self.assertIsNone(
            utils.check_allowed_fields(['network_interface']))

    def test_check_allowed_fields_network_interface_fail(self, mock_request):
        mock_request.version.minor = 19
        self.assertRaises(
            exception.NotAcceptable,
            utils.check_allowed_fields,
            ['network_interface'])

    def test_check_allowed_fields_resource_class(self, mock_request):
        mock_request.version.minor = 21
        self.assertIsNone(
            utils.check_allowed_fields(['resource_class']))

    def test_check_allowed_fields_resource_class_fail(self, mock_request):
        mock_request.version.minor = 20
        self.assertRaises(
            exception.NotAcceptable,
            utils.check_allowed_fields,
            ['resource_class'])

    def test_check_allowed_fields_rescue_interface_fail(self, mock_request):
        mock_request.version.minor = 31
        self.assertRaises(
            exception.NotAcceptable,
            utils.check_allowed_fields,
            ['rescue_interface'])

    def test_check_allowed_portgroup_fields_mode_properties(self,
                                                            mock_request):
        mock_request.version.minor = 26
        self.assertIsNone(
            utils.check_allowed_portgroup_fields(['mode']))
        self.assertIsNone(
            utils.check_allowed_portgroup_fields(['properties']))

    def test_check_allowed_portgroup_fields_mode_properties_fail(self,
                                                                 mock_request):
        mock_request.version.minor = 25
        self.assertRaises(
            exception.NotAcceptable,
            utils.check_allowed_portgroup_fields,
            ['mode'])
        self.assertRaises(
            exception.NotAcceptable,
            utils.check_allowed_portgroup_fields,
            ['properties'])

    def test_check_allow_specify_driver(self, mock_request):
        mock_request.version.minor = 16
        self.assertIsNone(utils.check_allow_specify_driver(['fake']))

    def test_check_allow_specify_driver_fail(self, mock_request):
        mock_request.version.minor = 15
        self.assertRaises(exception.NotAcceptable,
                          utils.check_allow_specify_driver, ['fake'])

    def test_check_allow_specify_resource_class(self, mock_request):
        mock_request.version.minor = 21
        self.assertIsNone(utils.check_allow_specify_resource_class(['foo']))

    def test_check_allow_specify_resource_class_fail(self, mock_request):
        mock_request.version.minor = 20
        self.assertRaises(exception.NotAcceptable,
                          utils.check_allow_specify_resource_class, ['foo'])

    def test_check_allow_filter_driver_type(self, mock_request):
        mock_request.version.minor = 30
        self.assertIsNone(utils.check_allow_filter_driver_type('classic'))

    def test_check_allow_filter_driver_type_none(self, mock_request):
        mock_request.version.minor = 29
        self.assertIsNone(utils.check_allow_filter_driver_type(None))

    def test_check_allow_filter_driver_type_fail(self, mock_request):
        mock_request.version.minor = 29
        self.assertRaises(exception.NotAcceptable,
                          utils.check_allow_filter_driver_type, 'classic')

    def test_check_allow_filter_by_conductor_group(self, mock_request):
        mock_request.version.minor = 46
        self.assertIsNone(utils.check_allow_filter_by_conductor_group('foo'))

    def test_check_allow_filter_by_conductor_group_none(self, mock_request):
        mock_request.version.minor = 46
        self.assertIsNone(utils.check_allow_filter_by_conductor_group(None))

    def test_check_allow_filter_by_conductor_group_fail(self, mock_request):
        mock_request.version.minor = 45
        self.assertRaises(exception.NotAcceptable,
                          utils.check_allow_filter_by_conductor_group, 'foo')

    def test_check_allow_driver_detail(self, mock_request):
        mock_request.version.minor = 30
        self.assertIsNone(utils.check_allow_driver_detail(True))

    def test_check_allow_driver_detail_false(self, mock_request):
        mock_request.version.minor = 30
        self.assertIsNone(utils.check_allow_driver_detail(False))

    def test_check_allow_driver_detail_none(self, mock_request):
        mock_request.version.minor = 29
        self.assertIsNone(utils.check_allow_driver_detail(None))

    def test_check_allow_driver_detail_fail(self, mock_request):
        mock_request.version.minor = 29
        self.assertRaises(exception.NotAcceptable,
                          utils.check_allow_driver_detail, True)

    def test_check_allow_manage_verbs(self, mock_request):
        mock_request.version.minor = 4
        utils.check_allow_management_verbs('manage')

    def test_check_allow_manage_verbs_fail(self, mock_request):
        mock_request.version.minor = 3
        self.assertRaises(exception.NotAcceptable,
                          utils.check_allow_management_verbs, 'manage')

    def test_check_allow_provide_verbs(self, mock_request):
        mock_request.version.minor = 4
        utils.check_allow_management_verbs('provide')

    def test_check_allow_provide_verbs_fail(self, mock_request):
        mock_request.version.minor = 3
        self.assertRaises(exception.NotAcceptable,
                          utils.check_allow_management_verbs, 'provide')

    def test_check_allow_inspect_verbs(self, mock_request):
        mock_request.version.minor = 6
        utils.check_allow_management_verbs('inspect')

    def test_check_allow_inspect_verbs_fail(self, mock_request):
        mock_request.version.minor = 5
        self.assertRaises(exception.NotAcceptable,
                          utils.check_allow_management_verbs, 'inspect')

    def test_check_allow_abort_verbs(self, mock_request):
        mock_request.version.minor = 13
        utils.check_allow_management_verbs('abort')

    def test_check_allow_abort_verbs_fail(self, mock_request):
        mock_request.version.minor = 12
        self.assertRaises(exception.NotAcceptable,
                          utils.check_allow_management_verbs, 'abort')

    def test_check_allow_clean_verbs(self, mock_request):
        mock_request.version.minor = 15
        utils.check_allow_management_verbs('clean')

    def test_check_allow_clean_verbs_fail(self, mock_request):
        mock_request.version.minor = 14
        self.assertRaises(exception.NotAcceptable,
                          utils.check_allow_management_verbs, 'clean')

    def test_check_allow_deploy_verbs(self, mock_request):
        mock_request.version.minor = 73
        utils.check_allow_management_verbs('deploy')

    def test_check_allow_deploy_verbs_fail(self, mock_request):
        mock_request.version.minor = 72
        self.assertRaises(exception.NotAcceptable,
                          utils.check_allow_management_verbs, 'deploy')

    def test_check_allow_undeploy_verbs(self, mock_request):
        mock_request.version.minor = 73
        utils.check_allow_management_verbs('undeploy')

    def test_check_allow_undeploy_verbs_fail(self, mock_request):
        mock_request.version.minor = 72
        self.assertFalse(
            utils.check_allow_management_verbs('undeploy'))

    def test_check_allow_unknown_verbs(self, mock_request):
        utils.check_allow_management_verbs('rebuild')

    def test_allow_inject_nmi(self, mock_request):
        mock_request.version.minor = 29
        self.assertTrue(utils.allow_inject_nmi())
        mock_request.version.minor = 28
        self.assertFalse(utils.allow_inject_nmi())

    def test_allow_links_node_states_and_driver_properties(self, mock_request):
        mock_request.version.minor = 14
        self.assertTrue(utils.allow_links_node_states_and_driver_properties())
        mock_request.version.minor = 10
        self.assertFalse(utils.allow_links_node_states_and_driver_properties())

    def test_check_allow_adopt_verbs_fail(self, mock_request):
        mock_request.version.minor = 16
        self.assertRaises(exception.NotAcceptable,
                          utils.check_allow_management_verbs, 'adopt')

    def test_check_allow_adopt_verbs(self, mock_request):
        mock_request.version.minor = 17
        utils.check_allow_management_verbs('adopt')

    def test_allow_port_internal_info(self, mock_request):
        mock_request.version.minor = 18
        self.assertTrue(utils.allow_port_internal_info())
        mock_request.version.minor = 17
        self.assertFalse(utils.allow_port_internal_info())

    def test_allow_port_advanced_net_fields(self, mock_request):
        mock_request.version.minor = 19
        self.assertTrue(utils.allow_port_advanced_net_fields())
        mock_request.version.minor = 18
        self.assertFalse(utils.allow_port_advanced_net_fields())

    def test_allow_ramdisk_endpoints(self, mock_request):
        mock_request.version.minor = 22
        self.assertTrue(utils.allow_ramdisk_endpoints())
        mock_request.version.minor = 21
        self.assertFalse(utils.allow_ramdisk_endpoints())

    def test_allow_portgroups(self, mock_request):
        mock_request.version.minor = 23
        self.assertTrue(utils.allow_portgroups())
        mock_request.version.minor = 22
        self.assertFalse(utils.allow_portgroups())

    def test_allow_portgroups_subcontrollers(self, mock_request):
        mock_request.version.minor = 24
        self.assertTrue(utils.allow_portgroups_subcontrollers())
        mock_request.version.minor = 23
        self.assertFalse(utils.allow_portgroups_subcontrollers())

    def test_allow_remove_chassis_uuid(self, mock_request):
        mock_request.version.minor = 25
        self.assertTrue(utils.allow_remove_chassis_uuid())
        mock_request.version.minor = 24
        self.assertFalse(utils.allow_remove_chassis_uuid())

    def test_allow_portgroup_mode_properties(self, mock_request):
        mock_request.version.minor = 26
        self.assertTrue(utils.allow_portgroup_mode_properties())
        mock_request.version.minor = 25
        self.assertFalse(utils.allow_portgroup_mode_properties())

    def test_allow_dynamic_drivers(self, mock_request):
        mock_request.version.minor = 30
        self.assertTrue(utils.allow_dynamic_drivers())
        mock_request.version.minor = 29
        self.assertFalse(utils.allow_dynamic_drivers())

    def test_allow_volume(self, mock_request):
        mock_request.version.minor = 32
        self.assertTrue(utils.allow_volume())
        mock_request.version.minor = 31
        self.assertFalse(utils.allow_volume())

    def test_allow_storage_interface(self, mock_request):
        mock_request.version.minor = 33
        self.assertTrue(utils.allow_storage_interface())
        mock_request.version.minor = 32
        self.assertFalse(utils.allow_storage_interface())

    def test_allow_traits(self, mock_request):
        mock_request.version.minor = 37
        self.assertTrue(utils.allow_traits())
        mock_request.version.minor = 36
        self.assertFalse(utils.allow_traits())

    @mock.patch.object(objects.Port, 'supports_physical_network',
                       autospec=True)
    def test_allow_port_physical_network_no_pin(self, mock_spn, mock_request):
        mock_spn.return_value = True
        mock_request.version.minor = 34
        self.assertTrue(utils.allow_port_physical_network())
        mock_request.version.minor = 33
        self.assertFalse(utils.allow_port_physical_network())

    @mock.patch.object(objects.Port, 'supports_physical_network',
                       autospec=True)
    def test_allow_port_physical_network_pin(self, mock_spn, mock_request):
        mock_spn.return_value = False
        mock_request.version.minor = 34
        self.assertFalse(utils.allow_port_physical_network())
        mock_request.version.minor = 33
        self.assertFalse(utils.allow_port_physical_network())

    def test_allow_node_rebuild_with_configdrive(self, mock_request):
        mock_request.version.minor = 35
        self.assertTrue(utils.allow_node_rebuild_with_configdrive())
        mock_request.version.minor = 34
        self.assertFalse(utils.allow_node_rebuild_with_configdrive())

    def test_allow_configdrive_vendor_data(self, mock_request):
        mock_request.version.minor = 59
        self.assertTrue(utils.allow_configdrive_vendor_data())
        mock_request.version.minor = 58
        self.assertFalse(utils.allow_configdrive_vendor_data())

    def test_check_allow_configdrive_fails(self, mock_request):
        mock_request.version.minor = 35
        self.assertRaises(exception.ClientSideError,
                          utils.check_allow_configdrive, states.DELETED,
                          "abcd")
        self.assertRaises(exception.ClientSideError,
                          utils.check_allow_configdrive, states.ACTIVE,
                          {'meta_data': {}})
        mock_request.version.minor = 34
        self.assertRaises(exception.ClientSideError,
                          utils.check_allow_configdrive, states.REBUILD,
                          "abcd")

    def test_check_allow_configdrive(self, mock_request):
        mock_request.version.minor = 35
        utils.check_allow_configdrive(states.ACTIVE, "abcd")
        utils.check_allow_configdrive(states.REBUILD, "abcd")
        mock_request.version.minor = 34
        utils.check_allow_configdrive(states.ACTIVE, "abcd")

    def test_check_allow_configdrive_as_dict(self, mock_request):
        mock_request.version.minor = 59
        utils.check_allow_configdrive(states.ACTIVE, {'meta_data': {}})
        utils.check_allow_configdrive(states.ACTIVE, {'meta_data': {},
                                                      'network_data': {},
                                                      'user_data': {},
                                                      'vendor_data': {}})
        utils.check_allow_configdrive(states.ACTIVE, {'user_data': 'foo'})
        utils.check_allow_configdrive(states.ACTIVE, {'user_data': ['foo']})

    def test_check_allow_configdrive_vendor_data_failed(self, mock_request):
        mock_request.version.minor = 58
        self.assertRaises(exception.ClientSideError,
                          utils.check_allow_configdrive,
                          states.ACTIVE,
                          {'meta_data': {},
                           'network_data': {},
                           'user_data': {},
                           'vendor_data': {}})

    def test_check_allow_configdrive_as_dict_invalid(self, mock_request):
        mock_request.version.minor = 59
        self.assertRaises(exception.ClientSideError,
                          utils.check_allow_configdrive, states.REBUILD,
                          {'foo': 'bar'})
        for key in ['meta_data', 'network_data']:
            self.assertRaises(exception.ClientSideError,
                              utils.check_allow_configdrive, states.REBUILD,
                              {key: 'a string'})
        for key in ['meta_data', 'network_data', 'user_data']:
            self.assertRaises(exception.ClientSideError,
                              utils.check_allow_configdrive, states.REBUILD,
                              {key: 42})

    def test_allow_rescue_interface(self, mock_request):
        mock_request.version.minor = 38
        self.assertTrue(utils.allow_rescue_interface())
        mock_request.version.minor = 37
        self.assertFalse(utils.allow_rescue_interface())

    def test_allow_inspect_abort(self, mock_request):
        mock_request.version.minor = 41
        self.assertTrue(utils.allow_inspect_abort())
        mock_request.version.minor = 40
        self.assertFalse(utils.allow_inspect_abort())

    def test_allow_port_is_smartnic(self, mock_request):
        mock_request.version.minor = 53
        self.assertTrue(utils.allow_port_is_smartnic())
        mock_request.version.minor = 52
        self.assertFalse(utils.allow_port_is_smartnic())

    def test_allow_deploy_templates(self, mock_request):
        mock_request.version.minor = 55
        self.assertTrue(utils.allow_deploy_templates())
        mock_request.version.minor = 54
        self.assertFalse(utils.allow_deploy_templates())

    def test_allow_agent_token(self, mock_request):
        mock_request.version.minor = 62
        self.assertTrue(utils.allow_agent_token())
        mock_request.version.minor = 61
        self.assertFalse(utils.allow_agent_token())

    def test_allow_deploy_steps(self, mock_request):
        mock_request.version.minor = 69
        self.assertTrue(utils.allow_deploy_steps())
        mock_request.version.minor = 68
        self.assertFalse(utils.allow_deploy_steps())

    def test_check_allow_deploy_steps(self, mock_request):
        mock_request.version.minor = 69
        utils.check_allow_deploy_steps(states.ACTIVE, {'a': 1})
        utils.check_allow_deploy_steps(states.REBUILD, {'a': 1})

    def test_check_allow_deploy_steps_empty(self, mock_request):
        utils.check_allow_deploy_steps(states.ACTIVE, None)

    def test_check_allow_deploy_steps_version_older(self, mock_request):
        mock_request.version.minor = 68
        self.assertRaises(exception.NotAcceptable,
                          utils.check_allow_deploy_steps,
                          states.ACTIVE, {'a': 1})

    def test_check_allow_deploy_steps_target_unsupported(self, mock_request):
        mock_request.version.minor = 69
        self.assertRaises(exception.ClientSideError,
                          utils.check_allow_deploy_steps,
                          states.MANAGEABLE, {'a': 1})


@mock.patch.object(api, 'request', spec_set=['context', 'version'])
class TestNodeIdent(base.TestCase):

    def setUp(self):
        super(TestNodeIdent, self).setUp()
        self.valid_name = 'my-host'
        self.valid_uuid = uuidutils.generate_uuid()
        self.invalid_name = 'Mr Plow'
        self.node = test_api_utils.post_get_test_node()

    def test_allow_node_logical_names_pre_name(self, mock_pecan_req):
        mock_pecan_req.version.minor = 1
        self.assertFalse(utils.allow_node_logical_names())

    def test_allow_node_logical_names_post_name(self, mock_pecan_req):
        mock_pecan_req.version.minor = 5
        self.assertTrue(utils.allow_node_logical_names())

    def test_is_valid_node_name(self, mock_pecan_req):
        mock_pecan_req.version.minor = 10
        self.assertTrue(utils.is_valid_node_name(self.valid_name))
        self.assertFalse(utils.is_valid_node_name(self.invalid_name))
        self.assertFalse(utils.is_valid_node_name(self.valid_uuid))

    @mock.patch.object(utils, 'allow_node_logical_names', autospec=True)
    @mock.patch.object(objects.Node, 'get_by_uuid', autospec=True)
    @mock.patch.object(objects.Node, 'get_by_name', autospec=True)
    def test_get_rpc_node_expect_uuid(self, mock_gbn, mock_gbu, mock_anln,
                                      mock_pr):
        mock_anln.return_value = True
        self.node['uuid'] = self.valid_uuid
        mock_gbu.return_value = self.node
        self.assertEqual(self.node, utils.get_rpc_node(self.valid_uuid))
        self.assertEqual(1, mock_gbu.call_count)
        self.assertEqual(0, mock_gbn.call_count)

    @mock.patch.object(utils, 'allow_node_logical_names', autospec=True)
    @mock.patch.object(objects.Node, 'get_by_uuid', autospec=True)
    @mock.patch.object(objects.Node, 'get_by_name', autospec=True)
    def test_get_rpc_node_expect_name(self, mock_gbn, mock_gbu, mock_anln,
                                      mock_pr):
        mock_pr.version.minor = 10
        mock_anln.return_value = True
        self.node['name'] = self.valid_name
        mock_gbn.return_value = self.node
        self.assertEqual(self.node, utils.get_rpc_node(self.valid_name))
        self.assertEqual(0, mock_gbu.call_count)
        self.assertEqual(1, mock_gbn.call_count)

    @mock.patch.object(utils, 'allow_node_logical_names', autospec=True)
    @mock.patch.object(objects.Node, 'get_by_uuid', autospec=True)
    @mock.patch.object(objects.Node, 'get_by_name', autospec=True)
    def test_get_rpc_node_invalid_name(self, mock_gbn, mock_gbu,
                                       mock_anln, mock_pr):
        mock_pr.version.minor = 10
        mock_anln.return_value = True
        self.assertRaises(exception.InvalidUuidOrName,
                          utils.get_rpc_node,
                          self.invalid_name)

    @mock.patch.object(utils, 'allow_node_logical_names', autospec=True)
    @mock.patch.object(objects.Node, 'get_by_uuid', autospec=True)
    @mock.patch.object(objects.Node, 'get_by_name', autospec=True)
    def test_get_rpc_node_by_uuid_no_logical_name(self, mock_gbn, mock_gbu,
                                                  mock_anln, mock_pr):
        # allow_node_logical_name() should have no effect
        mock_anln.return_value = False
        self.node['uuid'] = self.valid_uuid
        mock_gbu.return_value = self.node
        self.assertEqual(self.node, utils.get_rpc_node(self.valid_uuid))
        self.assertEqual(1, mock_gbu.call_count)
        self.assertEqual(0, mock_gbn.call_count)

    @mock.patch.object(utils, 'allow_node_logical_names', autospec=True)
    @mock.patch.object(objects.Node, 'get_by_uuid', autospec=True)
    @mock.patch.object(objects.Node, 'get_by_name', autospec=True)
    def test_get_rpc_node_by_name_no_logical_name(self, mock_gbn, mock_gbu,
                                                  mock_anln, mock_pr):
        mock_anln.return_value = False
        self.node['name'] = self.valid_name
        mock_gbn.return_value = self.node
        self.assertRaises(exception.NodeNotFound,
                          utils.get_rpc_node,
                          self.valid_name)

    @mock.patch.object(objects.Node, 'get_by_id', autospec=True)
    def test_populate_node_uuid(self, mock_gbi, mock_pr):
        port = obj_utils.get_test_port(self.context)
        node = obj_utils.get_test_node(self.context, id=port.node_id)
        mock_gbi.return_value = node

        # successful lookup
        d = {}
        utils.populate_node_uuid(port, d)
        self.assertEqual({
            'node_uuid': '1be26c0b-03f2-4d2e-ae87-c02d7f33c123'
        }, d)

        # not found, raise exception
        mock_gbi.side_effect = exception.NodeNotFound(node=port.node_id)
        d = {}
        self.assertRaises(exception.NodeNotFound,
                          utils.populate_node_uuid, port, d)

    @mock.patch.object(utils, 'check_owner_policy', autospec=True)
    @mock.patch.object(objects.Node, 'get_by_uuid', autospec=True)
    def test_replace_node_uuid_with_id(self, mock_gbu, mock_check, mock_pr):
        node = obj_utils.get_test_node(self.context, id=1)
        mock_gbu.return_value = node
        to_dict = {'node_uuid': self.valid_uuid}

        self.assertEqual(node, utils.replace_node_uuid_with_id(to_dict))
        self.assertEqual({'node_id': 1}, to_dict)
        mock_check.assert_called_once_with('node', 'baremetal:node:get',
                                           None, None,
                                           conceal_node=mock.ANY)

    @mock.patch.object(objects.Node, 'get_by_uuid', autospec=True)
    def test_replace_node_uuid_with_id_not_found(self, mock_gbu, mock_pr):
        to_dict = {'node_uuid': self.valid_uuid}
        mock_gbu.side_effect = exception.NodeNotFound(node=self.valid_uuid)

        e = self.assertRaises(exception.NodeNotFound,
                              utils.replace_node_uuid_with_id, to_dict)
        self.assertEqual(400, e.code)

    @mock.patch.object(objects.Node, 'get_by_id', autospec=True)
    def test_replace_node_id_with_uuid(self, mock_gbi, mock_pr):
        node = obj_utils.get_test_node(self.context, uuid=self.valid_uuid)
        mock_gbi.return_value = node
        to_dict = {'node_id': 1}

        self.assertEqual(node, utils.replace_node_id_with_uuid(to_dict))
        self.assertEqual({'node_uuid': self.valid_uuid}, to_dict)

    @mock.patch.object(objects.Node, 'get_by_id', autospec=True)
    def test_replace_node_id_with_uuid_not_found(self, mock_gbi, mock_pr):
        to_dict = {'node_id': 1}
        mock_gbi.side_effect = exception.NodeNotFound(node=1)

        e = self.assertRaises(exception.NodeNotFound,
                              utils.replace_node_id_with_uuid, to_dict)
        self.assertEqual(400, e.code)


class TestVendorPassthru(base.TestCase):

    def test_method_not_specified(self):
        self.assertRaises(exception.ClientSideError,
                          utils.vendor_passthru, 'fake-ident',
                          None, 'fake-topic', data='fake-data')

    @mock.patch.object(api, 'request',
                       spec_set=['method', 'context', 'rpcapi'])
    def _vendor_passthru(self, mock_request, async_call=True,
                         driver_passthru=False):
        return_value = {
            'return': 'SpongeBob',
            'async': async_call,
            'attach': False
        }
        mock_request.method = 'post'
        mock_request.context = 'fake-context'

        passthru_mock = None
        if driver_passthru:
            passthru_mock = mock_request.rpcapi.driver_vendor_passthru
        else:
            passthru_mock = mock_request.rpcapi.vendor_passthru
        passthru_mock.return_value = return_value

        response = utils.vendor_passthru('fake-ident', 'squarepants',
                                         'fake-topic', data='fake-data',
                                         driver_passthru=driver_passthru)

        passthru_mock.assert_called_once_with(
            'fake-context', 'fake-ident', 'squarepants', 'POST',
            'fake-data', 'fake-topic')
        self.assertIsInstance(response, utils.PassthruResponse)
        self.assertEqual('SpongeBob', response.obj)
        sc = http_client.ACCEPTED if async_call else http_client.OK
        self.assertEqual(sc, response.status_code)

    def test_vendor_passthru_async(self):
        self._vendor_passthru()

    def test_vendor_passthru_sync(self):
        self._vendor_passthru(async_call=False)

    def test_driver_vendor_passthru_async(self):
        self._vendor_passthru(driver_passthru=True)

    def test_driver_vendor_passthru_sync(self):
        self._vendor_passthru(async_call=False, driver_passthru=True)

    @mock.patch.object(api, 'request',
                       spec_set=['method', 'context', 'rpcapi'])
    def _test_vendor_passthru_attach(self, return_value, expct_return_value,
                                     mock_request):
        return_ = {'return': return_value, 'async': False, 'attach': True}
        mock_request.method = 'get'
        mock_request.context = 'fake-context'
        mock_request.rpcapi.driver_vendor_passthru.return_value = return_
        response = utils.vendor_passthru('fake-ident', 'bar',
                                         'fake-topic', data='fake-data',
                                         driver_passthru=True)
        mock_request.rpcapi.driver_vendor_passthru.assert_called_once_with(
            'fake-context', 'fake-ident', 'bar', 'GET',
            'fake-data', 'fake-topic')

        # Assert file was attached to the response object
        self.assertIsInstance(response.obj, io.BytesIO)
        self.assertEqual(expct_return_value, response.obj.read())
        # Assert response message is none
        self.assertIsInstance(response, utils.PassthruResponse)
        self.assertEqual(http_client.OK, response.status_code)

    def test_vendor_passthru_attach(self):
        self._test_vendor_passthru_attach('foo', b'foo')

    def test_vendor_passthru_attach_unicode_to_byte(self):
        self._test_vendor_passthru_attach(u'não', b'n\xc3\xa3o')

    def test_vendor_passthru_attach_byte_to_byte(self):
        self._test_vendor_passthru_attach(b'\x00\x01', b'\x00\x01')

    def test_get_controller_reserved_names(self):
        expected = ['maintenance', 'management', 'states',
                    'vendor_passthru', 'validate', 'detail']
        self.assertEqual(sorted(expected),
                         sorted(utils.get_controller_reserved_names(
                                api_node.NodesController)))

    @mock.patch.object(api, 'request', spec_set=["context"])
    @mock.patch.object(policy, 'authorize', spec=True)
    def test_check_policy(self, mock_authorize, mock_pr):
        fake_context = ironic_context.RequestContext()
        mock_pr.context = fake_context
        expected_target = dict(fake_context.to_policy_values())
        utils.check_policy('fake-policy')
        mock_authorize.assert_called_once_with('fake-policy', expected_target,
                                               fake_context)

    @mock.patch.object(api, 'request', spec_set=["context"])
    @mock.patch.object(policy, 'authorize', spec=True)
    def test_check_policy_forbidden(self, mock_authorize, mock_pr):
        mock_authorize.side_effect = exception.HTTPForbidden(resource='fake')
        self.assertRaises(exception.HTTPForbidden,
                          utils.check_policy, 'fake-policy')


class TestPortgroupIdent(base.TestCase):
    def setUp(self):
        super(TestPortgroupIdent, self).setUp()
        self.valid_name = 'my-portgroup'
        self.valid_uuid = uuidutils.generate_uuid()
        self.invalid_name = 'My Portgroup'
        self.portgroup = test_api_utils.post_get_test_portgroup()

    @mock.patch.object(api, 'request', spec_set=["context"])
    @mock.patch.object(objects.Portgroup, 'get_by_name', autospec=True)
    def test_get_rpc_portgroup_name(self, mock_gbn, mock_pr):
        mock_gbn.return_value = self.portgroup
        self.assertEqual(self.portgroup, utils.get_rpc_portgroup(
            self.valid_name))
        mock_gbn.assert_called_once_with(mock_pr.context, self.valid_name)

    @mock.patch.object(api, 'request', spec_set=["context"])
    @mock.patch.object(objects.Portgroup, 'get_by_uuid', autospec=True)
    def test_get_rpc_portgroup_uuid(self, mock_gbu, mock_pr):
        self.portgroup['uuid'] = self.valid_uuid
        mock_gbu.return_value = self.portgroup
        self.assertEqual(self.portgroup, utils.get_rpc_portgroup(
            self.valid_uuid))
        mock_gbu.assert_called_once_with(mock_pr.context, self.valid_uuid)

    def test_get_rpc_portgroup_invalid_name(self):
        self.assertRaises(exception.InvalidUuidOrName,
                          utils.get_rpc_portgroup,
                          self.invalid_name)


class TestCheckOwnerPolicy(base.TestCase):
    def setUp(self):
        super(TestCheckOwnerPolicy, self).setUp()
        self.valid_node_uuid = uuidutils.generate_uuid()
        self.node = test_api_utils.post_get_test_node()
        self.node['owner'] = '12345'
        self.node['lessee'] = '54321'

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    def test_check_owner_policy(
            self, mock_authorize, mock_pr
    ):
        fake_context = ironic_context.RequestContext()
        mock_pr.version.minor = 50
        mock_pr.context = fake_context
        expected_target = dict(fake_context.to_policy_values())
        expected_target['node.owner'] = '12345'
        expected_target['node.lessee'] = '54321'

        utils.check_owner_policy(
            'node', 'fake_policy', self.node['owner'], self.node['lessee']
        )
        mock_authorize.assert_called_once_with(
            'fake_policy', expected_target, fake_context)

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    def test_check_owner_policy_forbidden(
            self, mock_authorize, mock_pr
    ):
        mock_pr.version.minor = 50
        mock_pr.context.to_policy_values.return_value = {}
        mock_authorize.side_effect = exception.HTTPForbidden(resource='fake')

        self.assertRaises(
            exception.HTTPForbidden,
            utils.check_owner_policy,
            'node',
            'fake-policy',
            self.node
        )


class TestCheckNodePolicyAndRetrieve(base.TestCase):
    def setUp(self):
        super(TestCheckNodePolicyAndRetrieve, self).setUp()
        self.valid_node_uuid = uuidutils.generate_uuid()
        self.node = test_api_utils.post_get_test_node()
        self.node['owner'] = '12345'
        self.node['lessee'] = '54321'

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    @mock.patch.object(utils, 'get_rpc_node', autospec=True)
    @mock.patch.object(utils, 'get_rpc_node_with_suffix', autospec=True)
    def test_check_node_policy_and_retrieve(
            self, mock_grnws, mock_grn, mock_authorize, mock_pr
    ):
        fake_context = ironic_context.RequestContext()
        expected_target = dict(fake_context.to_policy_values())
        expected_target['node.owner'] = '12345'
        expected_target['node.lessee'] = '54321'
        mock_pr.context = fake_context

        mock_pr.version.minor = 50
        mock_grn.return_value = self.node

        rpc_node = utils.check_node_policy_and_retrieve(
            'fake_policy', self.valid_node_uuid
        )
        authorize_calls = [
            mock.call('baremetal:node:get', expected_target, fake_context),
            mock.call('fake_policy', expected_target, fake_context)]

        mock_grn.assert_called_once_with(self.valid_node_uuid)
        mock_grnws.assert_not_called()
        mock_authorize.assert_has_calls(authorize_calls)
        self.assertEqual(self.node, rpc_node)

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    @mock.patch.object(utils, 'get_rpc_node', autospec=True)
    @mock.patch.object(utils, 'get_rpc_node_with_suffix', autospec=True)
    def test_check_node_policy_and_retrieve_with_suffix(
            self, mock_grnws, mock_grn, mock_authorize, mock_pr
    ):
        fake_context = ironic_context.RequestContext()
        expected_target = fake_context.to_policy_values()
        expected_target['node.owner'] = '12345'
        expected_target['node.lessee'] = '54321'
        mock_pr.context = fake_context
        mock_pr.version.minor = 50
        mock_grnws.return_value = self.node

        rpc_node = utils.check_node_policy_and_retrieve(
            'fake_policy', self.valid_node_uuid, True
        )
        mock_grn.assert_not_called()
        mock_grnws.assert_called_once_with(self.valid_node_uuid)
        authorize_calls = [
            mock.call('baremetal:node:get', expected_target, fake_context),
            mock.call('fake_policy', expected_target, fake_context)]
        mock_authorize.assert_has_calls(authorize_calls)
        self.assertEqual(self.node, rpc_node)

    @mock.patch.object(api, 'request', spec_set=["context"])
    @mock.patch.object(policy, 'authorize', spec=True)
    @mock.patch.object(utils, 'get_rpc_node', autospec=True)
    def test_check_node_policy_and_retrieve_no_node_policy_notfound(
            self, mock_grn, mock_authorize, mock_pr
    ):
        mock_pr.context.to_policy_values.return_value = {}
        mock_authorize.side_effect = exception.HTTPForbidden(resource='fake')
        mock_grn.side_effect = exception.NodeNotFound(
            node=self.valid_node_uuid)

        self.assertRaises(
            exception.NodeNotFound,
            utils.check_node_policy_and_retrieve,
            'fake-policy',
            self.valid_node_uuid
        )

    @mock.patch.object(api, 'request', spec_set=["context"])
    @mock.patch.object(policy, 'authorize', spec=True)
    @mock.patch.object(utils, 'get_rpc_node', autospec=True)
    def test_check_node_policy_and_retrieve_no_node(
            self, mock_grn, mock_authorize, mock_pr
    ):
        mock_pr.context.to_policy_values.return_value = {}
        mock_grn.side_effect = exception.NodeNotFound(
            node=self.valid_node_uuid)

        self.assertRaises(
            exception.NodeNotFound,
            utils.check_node_policy_and_retrieve,
            'fake-policy',
            self.valid_node_uuid
        )

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    @mock.patch.object(utils, 'get_rpc_node', autospec=True)
    def test_check_node_policy_and_retrieve_policy_forbidden(
            self, mock_grn, mock_authorize, mock_pr
    ):
        mock_pr.version.minor = 50
        mock_pr.context.to_policy_values.return_value = {}
        mock_authorize.side_effect = exception.HTTPForbidden(resource='fake')
        mock_grn.return_value = self.node

        self.assertRaises(
            exception.NodeNotFound,
            utils.check_node_policy_and_retrieve,
            'fake-policy',
            self.valid_node_uuid
        )


class TestCheckAllocationPolicyAndRetrieve(base.TestCase):
    def setUp(self):
        super(TestCheckAllocationPolicyAndRetrieve, self).setUp()
        self.valid_allocation_uuid = uuidutils.generate_uuid()
        self.allocation = test_api_utils.allocation_post_data()
        self.allocation['owner'] = '12345'

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    @mock.patch.object(utils, 'get_rpc_allocation_with_suffix', autospec=True)
    def test_check_node_policy_and_retrieve(
            self, mock_graws, mock_authorize, mock_pr
    ):
        fake_context = ironic_context.RequestContext()
        expected_target = dict(fake_context.to_policy_values())
        expected_target['allocation.owner'] = '12345'
        mock_pr.version.minor = 60
        mock_pr.context = fake_context
        mock_graws.return_value = self.allocation

        rpc_allocation = utils.check_allocation_policy_and_retrieve(
            'fake_policy', self.valid_allocation_uuid
        )
        mock_graws.assert_called_once_with(self.valid_allocation_uuid)

        expected = [mock.call('baremetal:allocation:get',
                              expected_target, fake_context),
                    mock.call('fake_policy', expected_target,
                              fake_context)]
        mock_authorize.assert_has_calls(expected)
        self.assertEqual(self.allocation, rpc_allocation)

    @mock.patch.object(api, 'request', spec_set=["context"])
    @mock.patch.object(policy, 'authorize', spec=True)
    @mock.patch.object(utils, 'get_rpc_allocation_with_suffix', autospec=True)
    def test_check_alloc_policy_and_retrieve_no_alloc_policy_not_found(
            self, mock_graws, mock_authorize, mock_pr
    ):
        mock_pr.context.to_policy_values.return_value = {}
        mock_authorize.side_effect = exception.HTTPForbidden(resource='fake')
        mock_graws.side_effect = exception.AllocationNotFound(
            allocation=self.valid_allocation_uuid)

        self.assertRaises(
            exception.AllocationNotFound,
            utils.check_allocation_policy_and_retrieve,
            'fake-policy',
            self.valid_allocation_uuid
        )

    @mock.patch.object(api, 'request', spec_set=["context"])
    @mock.patch.object(policy, 'authorize', spec=True)
    @mock.patch.object(utils, 'get_rpc_allocation_with_suffix', autospec=True)
    def test_check_allocation_policy_and_retrieve_no_allocation(
            self, mock_graws, mock_authorize, mock_pr
    ):
        mock_pr.context.to_policy_values.return_value = {}
        mock_graws.side_effect = exception.AllocationNotFound(
            allocation=self.valid_allocation_uuid)

        self.assertRaises(
            exception.AllocationNotFound,
            utils.check_allocation_policy_and_retrieve,
            'fake-policy',
            self.valid_allocation_uuid
        )

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    @mock.patch.object(utils, 'get_rpc_allocation_with_suffix', autospec=True)
    def test_check_allocation_policy_and_retrieve_policy_not_found(
            self, mock_graws, mock_authorize, mock_pr
    ):
        mock_pr.version.minor = 50
        mock_pr.context.to_policy_values.return_value = {}
        mock_authorize.side_effect = exception.HTTPForbidden(resource='fake')
        mock_graws.return_value = self.allocation

        self.assertRaises(
            exception.AllocationNotFound,
            utils.check_allocation_policy_and_retrieve,
            'fake-policy',
            self.valid_allocation_uuid
        )


class TestCheckMultipleNodePoliciesAndRetrieve(base.TestCase):
    def setUp(self):
        super(TestCheckMultipleNodePoliciesAndRetrieve, self).setUp()
        self.valid_node_uuid = uuidutils.generate_uuid()
        self.node = test_api_utils.post_get_test_node()
        self.node['owner'] = '12345'
        self.node['lessee'] = '54321'

    @mock.patch.object(utils, 'check_node_policy_and_retrieve', autospec=True)
    @mock.patch.object(utils, 'check_owner_policy', autospec=True)
    def test_check_multiple_node_policies_and_retrieve(
            self, mock_cop, mock_cnpar
    ):
        mock_cnpar.return_value = self.node
        mock_cop.return_value = True

        rpc_node = utils.check_multiple_node_policies_and_retrieve(
            ['fake_policy_1', 'fake_policy_2'], self.valid_node_uuid
        )
        mock_cnpar.assert_called_once_with('fake_policy_1',
                                           self.valid_node_uuid, False)
        mock_cop.assert_called_once_with(
            'node', 'fake_policy_2', '12345', '54321')
        self.assertEqual(self.node, rpc_node)

    @mock.patch.object(utils, 'check_node_policy_and_retrieve', autospec=True)
    @mock.patch.object(utils, 'check_owner_policy', autospec=True)
    def test_check_multiple_node_policies_and_retrieve_first_fail(
            self, mock_cop, mock_cnpar
    ):
        mock_cnpar.side_effect = exception.HTTPForbidden(resource='fake')
        mock_cop.return_value = True

        self.assertRaises(
            exception.HTTPForbidden,
            utils.check_multiple_node_policies_and_retrieve,
            ['fake_policy_1', 'fake_policy_2'],
            self.valid_node_uuid
        )
        mock_cnpar.assert_called_once_with('fake_policy_1',
                                           self.valid_node_uuid, False)
        mock_cop.assert_not_called()

    @mock.patch.object(utils, 'check_node_policy_and_retrieve', autospec=True)
    @mock.patch.object(utils, 'check_owner_policy', autospec=True)
    def test_check_node_policy_and_retrieve_no_node(
            self, mock_cop, mock_cnpar
    ):
        mock_cnpar.return_value = self.node
        mock_cop.side_effect = exception.HTTPForbidden(resource='fake')

        self.assertRaises(
            exception.HTTPForbidden,
            utils.check_multiple_node_policies_and_retrieve,
            ['fake_policy_1', 'fake_policy_2'],
            self.valid_node_uuid
        )
        mock_cnpar.assert_called_once_with('fake_policy_1',
                                           self.valid_node_uuid, False)
        mock_cop.assert_called_once_with(
            'node', 'fake_policy_2', '12345', '54321')


class TestCheckListPolicy(base.TestCase):
    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    def test_check_list_policy(
            self, mock_authorize, mock_pr
    ):
        mock_pr.context.to_policy_values.return_value = {
            'project_id': '12345'
        }
        mock_pr.version.minor = 50

        owner = utils.check_list_policy('node')
        self.assertIsNone(owner)

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    def test_check_list_policy_with_owner(
            self, mock_authorize, mock_pr
    ):
        mock_pr.context.to_policy_values.return_value = {
            'project_id': '12345'
        }
        mock_pr.version.minor = 50

        owner = utils.check_list_policy('node', '12345')
        self.assertEqual(owner, '12345')

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    def test_check_list_policy_forbidden(
            self, mock_authorize, mock_pr
    ):
        def mock_authorize_function(rule, target, creds):
            raise exception.HTTPForbidden(resource='fake')
        mock_authorize.side_effect = mock_authorize_function
        mock_pr.context.to_policy_values.return_value = {
            'project_id': '12345'
        }
        mock_pr.version.minor = 50

        self.assertRaises(
            exception.HTTPForbidden,
            utils.check_list_policy,
            'node'
        )

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    def test_check_list_policy_forbidden_no_project(
            self, mock_authorize, mock_pr
    ):
        def mock_authorize_function(rule, target, creds):
            if rule == 'baremetal:node:list_all':
                raise exception.HTTPForbidden(resource='fake')
            return True
        mock_authorize.side_effect = mock_authorize_function
        mock_pr.context.to_policy_values.return_value = {}
        mock_pr.version.minor = 50

        self.assertRaises(
            exception.HTTPForbidden,
            utils.check_list_policy,
            'node'
        )

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    def test_check_list_policy_non_admin(
            self, mock_authorize, mock_pr
    ):
        def mock_authorize_function(rule, target, creds):
            if rule == 'baremetal:node:list_all':
                raise exception.HTTPForbidden(resource='fake')
            return True
        mock_authorize.side_effect = mock_authorize_function
        mock_pr.context.to_policy_values.return_value = {
            'project_id': '12345'
        }
        mock_pr.version.minor = 50

        owner = utils.check_list_policy('node')
        self.assertEqual(owner, '12345')

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    def test_check_list_policy_non_admin_owner_proj_mismatch(
            self, mock_authorize, mock_pr
    ):
        def mock_authorize_function(rule, target, creds):
            if rule == 'baremetal:node:list_all':
                raise exception.HTTPForbidden(resource='fake')
            return True
        mock_authorize.side_effect = mock_authorize_function
        mock_pr.context.to_policy_values.return_value = {
            'project_id': '12345'
        }
        mock_pr.version.minor = 50

        self.assertRaises(
            exception.HTTPForbidden,
            utils.check_list_policy,
            'node',
            '54321'
        )


class TestCheckPortPolicyAndRetrieve(base.TestCase):
    def setUp(self):
        super(TestCheckPortPolicyAndRetrieve, self).setUp()
        self.valid_port_uuid = uuidutils.generate_uuid()
        self.node = test_api_utils.post_get_test_node()
        self.node['owner'] = '12345'
        self.node['lessee'] = '54321'
        self.port = objects.Port(self.context, node_id=42)

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    @mock.patch.object(objects.Port, 'get_by_uuid', autospec=True)
    @mock.patch.object(objects.Node, 'get_by_id', autospec=True)
    def test_check_port_policy_and_retrieve(
            self, mock_ngbi, mock_pgbu, mock_authorize, mock_pr
    ):
        fake_context = ironic_context.RequestContext()
        expected_target = fake_context.to_policy_values()
        expected_target['node.owner'] = '12345'
        expected_target['node.lessee'] = '54321'
        mock_pr.context = fake_context
        mock_pr.version.minor = 50
        mock_pgbu.return_value = self.port
        mock_ngbi.return_value = self.node

        rpc_port, rpc_node = utils.check_port_policy_and_retrieve(
            'fake_policy', self.valid_port_uuid
        )
        mock_pgbu.assert_called_once_with(mock_pr.context,
                                          self.valid_port_uuid)
        mock_ngbi.assert_called_once_with(mock_pr.context, 42)
        expected = [
            mock.call('baremetal:node:get', expected_target, fake_context),
            mock.call('fake_policy', expected_target, fake_context)]

        mock_authorize.assert_has_calls(expected)

        self.assertEqual(self.port, rpc_port)
        self.assertEqual(self.node, rpc_node)

    @mock.patch.object(api, 'request', spec_set=["context"])
    @mock.patch.object(policy, 'authorize', spec=True)
    @mock.patch.object(objects.Port, 'get_by_uuid', autospec=True)
    def test_check_port_policy_and_retrieve_no_port_policy_forbidden(
            self, mock_pgbu, mock_authorize, mock_pr
    ):
        mock_pr.context.to_policy_values.return_value = {}
        mock_authorize.side_effect = exception.HTTPForbidden(resource='fake')
        mock_pgbu.side_effect = exception.PortNotFound(
            port=self.valid_port_uuid)

        self.assertRaises(
            exception.PortNotFound,
            utils.check_port_policy_and_retrieve,
            'fake-policy',
            self.valid_port_uuid
        )

    @mock.patch.object(api, 'request', spec_set=["context"])
    @mock.patch.object(policy, 'authorize', spec=True)
    @mock.patch.object(objects.Port, 'get_by_uuid', autospec=True)
    def test_check_port_policy_and_retrieve_no_port(
            self, mock_pgbu, mock_authorize, mock_pr
    ):
        mock_pr.context.to_policy_values.return_value = {}
        mock_pgbu.side_effect = exception.PortNotFound(
            port=self.valid_port_uuid)

        self.assertRaises(
            exception.PortNotFound,
            utils.check_port_policy_and_retrieve,
            'fake-policy',
            self.valid_port_uuid
        )

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    @mock.patch.object(objects.Port, 'get_by_uuid', autospec=True)
    @mock.patch.object(objects.Node, 'get_by_id', autospec=True)
    def test_check_port_policy_and_retrieve_policy_notfound(
            self, mock_ngbi, mock_pgbu, mock_authorize, mock_pr
    ):
        mock_pr.version.minor = 50
        mock_pr.context.to_policy_values.return_value = {}
        mock_authorize.side_effect = exception.HTTPForbidden(resource='fake')
        mock_pgbu.return_value = self.port
        mock_ngbi.return_value = self.node

        self.assertRaises(
            exception.PortNotFound,
            utils.check_port_policy_and_retrieve,
            'fake-policy',
            self.valid_port_uuid
        )


class TestCheckPortListPolicy(base.TestCase):
    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    def test_check_port_list_policy(
            self, mock_authorize, mock_pr
    ):
        mock_pr.context.to_policy_values.return_value = {
            'project_id': '12345'
        }
        mock_pr.version.minor = 50

        owner = utils.check_port_list_policy()
        self.assertIsNone(owner)

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    def test_check_port_list_policy_forbidden(
            self, mock_authorize, mock_pr
    ):
        def mock_authorize_function(rule, target, creds):
            raise exception.HTTPForbidden(resource='fake')
        mock_authorize.side_effect = mock_authorize_function
        mock_pr.context.to_policy_values.return_value = {
            'project_id': '12345'
        }
        mock_pr.version.minor = 50

        self.assertRaises(
            exception.HTTPForbidden,
            utils.check_port_list_policy,
        )

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    def test_check_port_list_policy_forbidden_no_project(
            self, mock_authorize, mock_pr
    ):
        def mock_authorize_function(rule, target, creds):
            if rule == 'baremetal:port:list_all':
                raise exception.HTTPForbidden(resource='fake')
            return True
        mock_authorize.side_effect = mock_authorize_function
        mock_pr.context.to_policy_values.return_value = {}
        mock_pr.version.minor = 50

        self.assertRaises(
            exception.HTTPForbidden,
            utils.check_port_list_policy,
        )

    @mock.patch.object(api, 'request', spec_set=["context", "version"])
    @mock.patch.object(policy, 'authorize', spec=True)
    def test_check_port_list_policy_non_admin(
            self, mock_authorize, mock_pr
    ):
        def mock_authorize_function(rule, target, creds):
            if rule == 'baremetal:port:list_all':
                raise exception.HTTPForbidden(resource='fake')
            return True
        mock_authorize.side_effect = mock_authorize_function
        mock_pr.context.to_policy_values.return_value = {
            'project_id': '12345'
        }
        mock_pr.version.minor = 50

        owner = utils.check_port_list_policy()
        self.assertEqual(owner, '12345')


class TestObjectToDict(base.TestCase):

    def setUp(self):
        super(TestObjectToDict, self).setUp()
        self.node = obj_utils.get_test_node(
            self.context,
            created_at=datetime.datetime(2000, 1, 1, 0, 0),
            updated_at=datetime.datetime(2001, 1, 1, 0, 0),
            inspection_started_at=datetime.datetime(2002, 1, 1, 0, 0),
            console_enabled=True)

        p = mock.patch.object(api, 'request', autospec=False)
        mock_req = p.start()
        mock_req.public_url = 'http://192.0.2.1:5050'
        self.addCleanup(p.stop)

    def test_no_args(self):
        self.assertEqual({
            'created_at': '2000-01-01T00:00:00+00:00',
            'updated_at': '2001-01-01T00:00:00+00:00',
            'uuid': '1be26c0b-03f2-4d2e-ae87-c02d7f33c123'
        }, utils.object_to_dict(self.node))

    def test_no_base_attributes(self):
        self.assertEqual({}, utils.object_to_dict(
            self.node,
            include_created_at=False,
            include_updated_at=False,
            include_uuid=False)
        )

    def test_fields(self):
        self.assertEqual({
            'conductor_group': '',
            'console_enabled': True,
            'created_at': '2000-01-01T00:00:00+00:00',
            'driver': 'fake-hardware',
            'inspection_finished_at': None,
            'inspection_started_at': '2002-01-01T00:00:00+00:00',
            'maintenance': False,
            'updated_at': '2001-01-01T00:00:00+00:00',
            'uuid': '1be26c0b-03f2-4d2e-ae87-c02d7f33c123'
        }, utils.object_to_dict(
            self.node,
            fields=[
                'conductor_group',
                'console_enabled',
                'driver',
                'inspection_finished_at',
                'inspection_started_at',
                'maintenance',
            ])
        )

    def test_links(self):
        self.assertEqual({
            'created_at': '2000-01-01T00:00:00+00:00',
            'links': [{
                'href': 'http://192.0.2.1:5050/v1/node/'
                        '1be26c0b-03f2-4d2e-ae87-c02d7f33c123',
                'rel': 'self'
            }, {
                'href': 'http://192.0.2.1:5050/node/'
                        '1be26c0b-03f2-4d2e-ae87-c02d7f33c123',
                'rel': 'bookmark'
            }],
            'updated_at': '2001-01-01T00:00:00+00:00',
            'uuid': '1be26c0b-03f2-4d2e-ae87-c02d7f33c123',
        }, utils.object_to_dict(self.node, link_resource='node'))

        self.assertEqual({
            'created_at': '2000-01-01T00:00:00+00:00',
            'links': [{
                'href': 'http://192.0.2.1:5050/v1/node/foo',
                'rel': 'self'
            }, {
                'href': 'http://192.0.2.1:5050/node/foo',
                'rel': 'bookmark'
            }],
            'updated_at': '2001-01-01T00:00:00+00:00',
            'uuid': '1be26c0b-03f2-4d2e-ae87-c02d7f33c123',
        }, utils.object_to_dict(
            self.node,
            link_resource='node',
            link_resource_args='foo'))


class TestLocalLinkValidation(base.TestCase):

    def test_local_link_connection_type(self):
        v = utils.LOCAL_LINK_VALIDATOR
        value = {'switch_id': '0a:1b:2c:3d:4e:5f',
                 'port_id': 'value2',
                 'switch_info': 'value3'}
        self.assertEqual(value, v('l', value))

    def test_local_link_connection_type_datapath_id(self):
        v = utils.LOCAL_LINK_VALIDATOR
        value = {'switch_id': '0000000000000000',
                 'port_id': 'value2',
                 'switch_info': 'value3'}
        self.assertEqual(value, v('l', value))

    def test_local_link_connection_type_not_mac_or_datapath_id(self):
        v = utils.LOCAL_LINK_VALIDATOR
        value = {'switch_id': 'badid',
                 'port_id': 'value2',
                 'switch_info': 'value3'}
        self.assertRaises(exception.InvalidSwitchID, v, 'l', value)

    def test_local_link_connection_type_invalid_key(self):
        v = utils.LOCAL_LINK_VALIDATOR
        value = {'switch_id': '0a:1b:2c:3d:4e:5f',
                 'port_id': 'value2',
                 'switch_info': 'value3',
                 'invalid_key': 'value'}
        self.assertRaisesRegex(
            exception.Invalid,
            'Additional properties are not allowed',
            v, 'l', value)

    def test_local_link_connection_type_missing_local_link_mandatory_key(self):
        v = utils.LOCAL_LINK_VALIDATOR
        value = {'switch_id': '0a:1b:2c:3d:4e:5f',
                 'switch_info': 'value3'}
        self.assertRaisesRegex(exception.Invalid, 'is a required property',
                               v, 'l', value)

    def test_local_link_connection_type_local_link_keys_mandatory(self):
        v = utils.LOCAL_LINK_VALIDATOR
        value = {'switch_id': '0a:1b:2c:3d:4e:5f',
                 'port_id': 'value2'}
        self.assertEqual(value, v('l', value))

    def test_local_link_connection_type_empty_value(self):
        v = utils.LOCAL_LINK_VALIDATOR
        value = {}
        self.assertEqual(value, v('l', value))

    def test_local_link_connection_type_smart_nic_keys_mandatory(self):
        v = utils.LOCAL_LINK_VALIDATOR
        vs = utils.LOCAL_LINK_SMART_NIC_VALIDATOR
        value = {'port_id': 'rep0-0',
                 'hostname': 'hostname'}
        self.assertEqual(value, vs('l', value))
        self.assertEqual(value, v('l', value))

    def test_local_link_connection_type_smart_nic_keys_with_optional(self):
        v = utils.LOCAL_LINK_VALIDATOR
        vs = utils.LOCAL_LINK_SMART_NIC_VALIDATOR
        value = {'port_id': 'rep0-0',
                 'hostname': 'hostname',
                 'switch_id': '0a:1b:2c:3d:4e:5f',
                 'switch_info': 'sw_info'}
        self.assertEqual(value, vs('l', value))
        self.assertEqual(value, v('l', value))

    def test_local_link_connection_type_smart_nic_keys_hostname_missing(self):
        v = utils.LOCAL_LINK_VALIDATOR
        vs = utils.LOCAL_LINK_SMART_NIC_VALIDATOR
        value = {'port_id': 'rep0-0'}
        self.assertRaises(exception.Invalid, vs, 'l', value)
        self.assertRaises(exception.Invalid, v, 'l', value)

    def test_local_link_connection_type_smart_nic_keys_port_id_missing(self):
        v = utils.LOCAL_LINK_VALIDATOR
        vs = utils.LOCAL_LINK_SMART_NIC_VALIDATOR
        value = {'hostname': 'hostname'}
        self.assertRaises(exception.Invalid, vs, 'l', value)
        self.assertRaises(exception.Invalid, v, 'l', value)

    def test_local_link_connection_net_type_unmanaged(self):
        v = utils.LOCAL_LINK_VALIDATOR
        value = {'network_type': 'unmanaged'}
        self.assertEqual(value, v('l', value))

    def test_local_link_connection_net_type_unmanaged_combine_ok(self):
        v = utils.LOCAL_LINK_VALIDATOR
        value = {'network_type': 'unmanaged',
                 'switch_id': '0a:1b:2c:3d:4e:5f',
                 'port_id': 'rep0-0'}
        self.assertEqual(value, v('l', value))

    def test_local_link_connection_net_type_invalid(self):
        v = utils.LOCAL_LINK_VALIDATOR
        value = {'network_type': 'invalid'}
        self.assertRaises(exception.Invalid, v, 'l', value)

    def test_local_link_connection_cant_set_only_physical(self):
        v = utils.LOCAL_LINK_VALIDATOR
        value = {'port_id': '42',
                 'vtep-physical-switch': 'jswitch',
                 'switch_id': '0a:1b:2c:3d:4e:5f'}
        self.assertRaisesRegex(
            exception.Invalid,
            'is a dependency of',
            v, 'l', value)

    def test_local_link_connection_cant_set_only_logical(self):
        v = utils.LOCAL_LINK_VALIDATOR
        value = {'port_id': '42',
                 'vtep-logical-switch': 'jswitch',
                 'switch_id': '0a:1b:2c:3d:4e:5f'}
        self.assertRaisesRegex(
            exception.Invalid,
            'is a dependency of',
            v, 'l', value
        )

    def test_local_link_connection_set_both_switches(self):
        v = utils.LOCAL_LINK_VALIDATOR
        value = {'port_id': '42',
                 'vtep-logical-switch': 'lswitch',
                 'vtep-physical-switch': 'pswitch'}
        self.assertEqual(value, v('l', value))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""API request/response validating middleware."""

import functools
import inspect
import typing as ty

from oslo_serialization import jsonutils
from webob import exc as webob_exc

from ironic import api
from ironic.api.validation import validators
from ironic.common.i18n import _


def api_version(
    min_version: ty.Optional[int],
    max_version: ty.Optional[int] = None,
    message: ty.Optional[str] = None,
    exception_class: ty.Type[webob_exc.HTTPException] = webob_exc.HTTPNotFound,
):
    """Decorator for marking lower and upper bounds on API methods.

    :param min_version: An integer representing the minimum API version that
        the API is available under.
    :param max_version: An integer representing the maximum API version that
        the API is available under.
    :param message: A message to return if the API is not supported.
    :param exception_class: The exception class to raise if the API version is
        not supported (default is HTTPNotFound).
    """

    # Ensure the provided status code is valid for the given exception class
    assert isinstance(
        exception_class,
        type(webob_exc.HTTPException)
    ), (
        "Invalid exception class provided, must be a "
        "subclass of webob_exc.HTTPException."
    )

    def add_validator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Version checks
            if (
                min_version and not api.request.version.minor >= min_version
            ) or (
                max_version and not api.request.version.minor <= max_version
            ):
                # Raise provided exception with localized message
                raise exception_class(
                    detail=_(
                        message
                        or 'The API is not supported for this version'
                    )
                )

            return func(*args, **kwargs)

        wrapper.min_version = min_version
        wrapper.max_version = max_version

        return wrapper

    return add_validator


class Schemas:
    """A microversion-aware schema container.

    Allow definition and retrieval of schemas on a microversion-aware basis.
    """

    def __init__(self) -> None:
        self._schemas: list[
            tuple[dict[str, object], ty.Optional[int], ty.Optional[int]]
        ] = []

    def add_schema(
        self,
        schema: tuple[dict[str, object]],
        min_version: ty.Optional[int],
        max_version: ty.Optional[int],
    ) -> None:
        self._schemas.append((schema, min_version, max_version))

    def __call__(self) -> ty.Optional[dict[str, object]]:
        for schema, min_version, max_version in self._schemas:
            if (
                min_version and not api.request.version.minor >= min_version
            ) or (
                max_version and not api.request.version.minor <= max_version
            ):
                continue

            return schema

        return None


def _schema_validator(
    schema: ty.Dict[str, ty.Any],
    target: ty.Dict[str, ty.Any],
    min_version: ty.Optional[int],
    max_version: ty.Optional[int],
    is_body: bool = True,
):
    """A helper method to execute JSON Schema Validation.

    This method checks the request version whether matches the specified
    ``max_version`` and ``min_version``. If the version range matches the
    request, we validate ``schema`` against ``target``. A failure will result
    in ``ValidationError`` being raised.

    :param schema: The JSON Schema schema used to validate the target.
    :param target: The target to be validated by the schema.
    :param min_version: An integer indicating the minimum API version
        ``schema`` applies against.
    :param max_version: An integer indicating the maximum API version
        ``schema`` applies against.
    :param args: Positional arguments which passed into original method.
    :param kwargs: Keyword arguments which passed into original method.
    :param is_body: Whether ``target`` is a HTTP request body or not.
    :returns: None.
    :raises: ``ValidationError`` if validation fails.
    """
    # Only validate against the schema if it lies within
    # the version range specified. Note that if both min
    # and max are not specified the validator will always
    # be run.
    if (
        (min_version and api.request.version.minor < min_version)
        or (max_version and api.request.version.minor > max_version)
    ):
        return

    schema_validator = validators.SchemaValidator(schema, is_body=is_body)
    schema_validator.validate(target)


def _extract_parameters(function):
    sig = inspect.signature(function)
    params = []

    for param in sig.parameters.values():
        if param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
            if param.name == 'self':  # skip validating self
                continue

            params.append(param)
    return params


def request_parameter_schema(
    schema: ty.Dict[str, ty.Any],
    min_version: ty.Optional[int] = None,
    max_version: ty.Optional[int] = None,
):
    """Decorator for registering a request parameter schema on API methods.

    ``schema`` will be used for validating request parameters just before
    the API method is executed.

    :param schema: The JSON Schema schema used to validate the target.
    :param min_version: An integer indicating the minimum API version
        ``schema`` applies against.
    :param max_version: An integer indicating the maximum API version
        ``schema`` applies against.
    """

    def add_validator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # we need to convert positional arguments to a dict mapping token
            # name to value so that we have a reference to compare against
            parameters = _extract_parameters(func)
            if func.__name__ in ('patch', 'post'):
                # if this a create or update method, we need to ignore the
                # request body parameter
                parameters = parameters[:-1]

            parameters = {
                p.name: args[i + 1] for i, p in enumerate(parameters)
                if p.name != '_' and p.default is p.empty
            }
            _schema_validator(
                schema,
                parameters,
                min_version,
                max_version,
                is_body=True,
            )
            return func(*args, **kwargs)

        if not hasattr(wrapper, 'request_parameter_schemas'):
            wrapper.request_parameter_schemas = Schemas()

        wrapper.request_parameter_schemas .add_schema(
            schema, min_version, max_version
        )

        return wrapper

    return add_validator


def request_query_schema(
    schema: ty.Dict[str, ty.Any],
    min_version: ty.Optional[int] = None,
    max_version: ty.Optional[int] = None,
):
    """Decorator for registering a request query string schema on API methods.

    ``schema`` will be used for validating request query strings just before
    the API method is executed.

    :param schema: The JSON Schema schema used to validate the target.
    :param min_version: An integer indicating the minimum API version
        ``schema`` applies against.
    :param max_version: An integer indicating the maximum API version
        ``schema`` applies against.
    """

    def add_validator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _schema_validator(
                schema,
                kwargs,
                min_version,
                max_version,
                is_body=True,
            )
            return func(*args, **kwargs)

        if not hasattr(wrapper, 'request_query_schemas'):
            wrapper.request_query_schemas = Schemas()

        wrapper.request_query_schemas .add_schema(
            schema, min_version, max_version
        )

        return wrapper

    return add_validator


def request_body_schema(
    schema: ty.Dict[str, ty.Any],
    min_version: ty.Optional[str] = None,
    max_version: ty.Optional[str] = None,
):
    """Decorator for registering a request body schema on API methods.

    ``schema`` will be used for validating the request body just before the API
    method is executed.

    :param schema: The JSON Schema schema used to validate the target.
    :param min_version: A string indicating the minimum API version ``schema``
        applies against.
    :param max_version: A string indicating the maximum API version ``schema``
        applies against.
    """

    def add_validator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            parameters = _extract_parameters(func)
            if not parameters:
                # TODO(stephenfin): this would be a better check if we
                # distinguished between 'create' operations (which should have
                # at least one parameter, the body) and "update" operations
                # (which should have at least two, the IDs and the body)
                raise RuntimeError(
                    'The ironic.api.method.body decorator must be placed '
                    'outside the validation helpers to ensure it runs first.'
                )

            _schema_validator(
                schema,
                # The body argument will always be the last one
                kwargs[parameters[-1].name],
                min_version,
                max_version,
                is_body=True,
            )
            return func(*args, **kwargs)

        if not hasattr(wrapper, 'request_body_schemas'):
            wrapper.request_body_schemas = Schemas()

        wrapper.request_body_schemas .add_schema(
            schema, min_version, max_version
        )

        return wrapper

    return add_validator


def response_body_schema(
    schema: ty.Dict[str, ty.Any],
    min_version: ty.Optional[str] = None,
    max_version: ty.Optional[str] = None,
):
    """Decorator for registering a response body schema on API methods.

    ``schema`` will be used for validating the response body just after the API
    method is executed.

    :param schema: The JSON Schema schema used to validate the target.
    :param min_version: A string indicating the minimum API version ``schema``
        applies against.
    :param max_version: A string indicating the maximum API version ``schema``
        applies against.
    """

    def add_validator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            response = func(*args, **kwargs)

            # FIXME(stephenfin): How is ironic/pecan doing jsonification? The
            # below will fail on e.g. date-time fields

            # NOTE(stephenfin): If our response is an object, we need to
            # serialize and deserialize to convert e.g. date-time to strings
            _body = jsonutils.dumps(response)

            if _body == b'':
                body = None
            else:
                body = jsonutils.loads(_body)

            _schema_validator(
                schema,
                body,
                min_version,
                max_version,
                is_body=True,
            )
            return response

        if not hasattr(wrapper, 'response_body_schemas'):
            wrapper.response_body_schemas = Schemas()

        wrapper.response_body_schemas .add_schema(
            schema, min_version, max_version
        )

        return wrapper

    return add_validator



# coding=utf-8

# Copyright 2013 Hewlett-Packard Development Company, L.P.
# Copyright 2013 International Business Machines Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Test class for Ironic ManagerService."""

from unittest import mock

from oslo_config import cfg

from ironic.common import exception
from ironic.common import states
from ironic.conductor import task_manager
from ironic.conductor import utils as conductor_utils
from ironic.conductor import verify
from ironic import objects
from ironic.tests.unit.conductor import mgr_utils
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils

CONF = cfg.CONF


@mgr_utils.mock_record_keepalive
class DoNodeVerifyTestCase(mgr_utils.ServiceSetUpMixin, db_base.DbTestCase):
    @mock.patch.object(conductor_utils, 'node_cache_vendor', autospec=True)
    @mock.patch('ironic.objects.node.NodeCorrectedPowerStateNotification',
                autospec=True)
    @mock.patch('ironic.drivers.modules.fake.FakePower.get_power_state',
                autospec=True)
    @mock.patch('ironic.drivers.modules.fake.FakePower.validate',
                autospec=True)
    def test__do_node_verify(self, mock_validate, mock_get_power_state,
                             mock_notif, mock_cache_vendor):
        self._start_service()
        mock_get_power_state.return_value = states.POWER_OFF
        # Required for exception handling
        mock_notif.__name__ = 'NodeCorrectedPowerStateNotification'
        node = obj_utils.create_test_node(
            self.context, driver='fake-hardware',
            provision_state=states.VERIFYING,
            target_provision_state=states.MANAGEABLE,
            last_error=None,
            power_state=states.NOSTATE)

        with task_manager.acquire(
                self.context, node['id'], shared=False) as task:
            verify.do_node_verify(task)

        self._stop_service()

        # 1 notification should be sent -
        # baremetal.node.power_state_corrected.success
        mock_notif.assert_called_once_with(publisher=mock.ANY,
                                           event_type=mock.ANY,
                                           level=mock.ANY,
                                           payload=mock.ANY)
        mock_notif.return_value.emit.assert_called_once_with(mock.ANY)

        node.refresh()

        mock_validate.assert_called_once_with(mock.ANY, task)
        mock_get_power_state.assert_called_once_with(mock.ANY, task)

        self.assertEqual(states.MANAGEABLE, node.provision_state)
        self.assertIsNone(node.target_provision_state)
        self.assertIsNone(node.last_error)
        self.assertEqual(states.POWER_OFF, node.power_state)

    @mock.patch('ironic.drivers.modules.fake.FakePower.get_power_state',
                autospec=True)
    @mock.patch('ironic.drivers.modules.fake.FakePower.validate',
                autospec=True)
    def test__do_node_verify_validation_fails(self, mock_validate,
                                              mock_get_power_state):
        self._start_service()
        node = obj_utils.create_test_node(
            self.context, driver='fake-hardware',
            provision_state=states.VERIFYING,
            target_provision_state=states.MANAGEABLE,
            last_error=None,
            power_state=states.NOSTATE)

        mock_validate.side_effect = RuntimeError("boom")

        with task_manager.acquire(
                self.context, node['id'], shared=False) as task:
            verify.do_node_verify(task)

        self._stop_service()
        node.refresh()

        mock_validate.assert_called_once_with(mock.ANY, task)

        self.assertEqual(states.ENROLL, node.provision_state)
        self.assertIsNone(node.target_provision_state)
        self.assertTrue(node.last_error)
        self.assertFalse(mock_get_power_state.called)

    @mock.patch('ironic.drivers.modules.fake.FakePower.get_power_state',
                autospec=True)
    @mock.patch('ironic.drivers.modules.fake.FakePower.validate',
                autospec=True)
    def test__do_node_verify_get_state_fails(self, mock_validate,
                                             mock_get_power_state):
        self._start_service()
        node = obj_utils.create_test_node(
            self.context, driver='fake-hardware',
            provision_state=states.VERIFYING,
            target_provision_state=states.MANAGEABLE,
            last_error=None,
            power_state=states.NOSTATE)

        mock_get_power_state.side_effect = RuntimeError("boom")

        with task_manager.acquire(
                self.context, node['id'], shared=False) as task:
            verify.do_node_verify(task)

        self._stop_service()
        node.refresh()

        mock_get_power_state.assert_called_once_with(mock.ANY, task)

        self.assertEqual(states.ENROLL, node.provision_state)
        self.assertIsNone(node.target_provision_state)
        self.assertTrue(node.last_error)
        history = objects.NodeHistory.list_by_node_id(self.context,
                                                      node.id)
        entry = history[0]
        self.assertEqual('verify', entry['event_type'])
        self.assertEqual('ERROR', entry['severity'])

    @mock.patch.object(conductor_utils, 'LOG', autospec=True)
    @mock.patch('ironic.drivers.modules.fake.FakePower.validate',
                autospec=True)
    @mock.patch('ironic.drivers.modules.fake.FakeBIOS.cache_bios_settings',
                autospec=True)
    def _test__do_node_cache_bios(self, mock_bios, mock_validate,
                                  mock_log,
                                  enable_unsupported=False,
                                  enable_exception=False):
        if enable_unsupported:
            mock_bios.side_effect = exception.UnsupportedDriverExtension('')
        elif enable_exception:
            mock_bios.side_effect = exception.IronicException('test')
        node = obj_utils.create_test_node(
            self.context, driver='fake-hardware',
            provision_state=states.VERIFYING,
            target_provision_state=states.MANAGEABLE)
        with task_manager.acquire(
                self.context, node.uuid, shared=False) as task:
            verify.do_node_verify(task)
            mock_bios.assert_called_once_with(mock.ANY, task)
            mock_validate.assert_called_once_with(mock.ANY, task)
            if enable_exception:
                mock_log.exception.assert_called_once_with(
                    'Caching of bios settings failed on node {}.'
                    .format(node.uuid))

    def test__do_node_cache_bios(self):
        self._test__do_node_cache_bios()

    def test__do_node_cache_bios_exception(self):
        self._test__do_node_cache_bios(enable_exception=True)

    def test__do_node_cache_bios_unsupported(self):
        self._test__do_node_cache_bios(enable_unsupported=True)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
DRAC inspection interface
"""

from ironic.common import boot_modes
from ironic.drivers.modules.drac import utils as drac_utils
from ironic.drivers.modules import inspect_utils
from ironic.drivers.modules.redfish import inspect as redfish_inspect
from ironic.drivers.modules.redfish import utils as redfish_utils


_PXE_DEV_ENABLED_INTERFACES = [('PxeDev1EnDis', 'PxeDev1Interface'),
                               ('PxeDev2EnDis', 'PxeDev2Interface'),
                               ('PxeDev3EnDis', 'PxeDev3Interface'),
                               ('PxeDev4EnDis', 'PxeDev4Interface')]
_BIOS_ENABLED_VALUE = 'Enabled'


class DracRedfishInspect(redfish_inspect.RedfishInspect):
    """iDRAC Redfish interface for inspection-related actions."""

    def inspect_hardware(self, task):
        """Inspect hardware to get the hardware properties.

        Inspects hardware to get the essential properties.
        It fails if any of the essential properties
        are not received from the node.

        :param task: a TaskManager instance.
        :raises: HardwareInspectionFailure if essential properties
                 could not be retrieved successfully.
        :returns: The resulting state of inspection.

        """
        # Ensure we create a port for every NIC port found for consistency
        # with our previous WSMAN inspect behavior and to work around a bug
        # in some versions of the firmware where the port state is not being
        # reported correctly.

        ethernet_interfaces_mac = list(self._get_mac_address(task).values())
        inspect_utils.create_ports_if_not_exist(task, ethernet_interfaces_mac)
        return super(DracRedfishInspect, self).inspect_hardware(task)

    def _get_mac_address(self, task):
        """Get a list of MAC addresses

        :param task: a TaskManager instance.
        :returns: a mapping of interface identities to MAC addresses.
        """
        system = redfish_utils.get_system(task.node)
        # Get dictionary of ethernet interfaces
        if system.ethernet_interfaces and system.ethernet_interfaces.summary:
            ethernet_interfaces = system.ethernet_interfaces.get_members()
            ethernet_interfaces_mac = {
                interface.identity: interface.mac_address
                for interface in ethernet_interfaces}
            return ethernet_interfaces_mac
        else:
            return {}

    def _get_pxe_port_macs(self, task):
        """Get a list of PXE port MAC addresses.

        :param task: a TaskManager instance.
        :returns: Returns list of PXE port MAC addresses.
        """
        system = redfish_utils.get_system(task.node)
        ethernet_interfaces_mac = self._get_mac_address(task)
        pxe_port_macs = []

        if system.boot.mode == boot_modes.UEFI:
            # When a server is in UEFI boot mode, the PXE NIC ports are
            # stored in the PxeDevXEnDis and PxeDevXInterface BIOS
            # settings. Get the PXE NIC ports from these settings and
            # their MAC addresses.
            for param, nic in _PXE_DEV_ENABLED_INTERFACES:
                if system.bios.attributes[param] == _BIOS_ENABLED_VALUE:
                    nic_id = system.bios.attributes[nic]
                    # Get MAC address of the given nic_id
                    mac_address = ethernet_interfaces_mac[nic_id]
                    pxe_port_macs.append(mac_address)
        elif system.boot.mode == boot_modes.LEGACY_BIOS:
            # When a server is in BIOS boot mode, whether or not a
            # NIC port is set to PXE boot is stored on the NIC port
            # itself internally to the BMC. Getting this information
            # requires using an OEM extension to export the system
            # configuration, as the redfish standard does not specify
            # how to get it, and Dell does not have OEM redfish calls
            # to selectively retrieve it at this time.
            # Get instance of Sushy OEM manager object
            pxe_port_macs_list = drac_utils.execute_oem_manager_method(
                task, 'get PXE port MAC addresses',
                lambda m: m.get_pxe_port_macs_bios(ethernet_interfaces_mac))
            pxe_port_macs = [mac for mac in pxe_port_macs_list]

        return pxe_port_macs



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""oslo.i18n integration module.

See https://docs.openstack.org/oslo.i18n/latest/user/

"""

try:
    import oslo_i18n as i18n
except ImportError:
    def _(msg):
        return msg

    def install(domain):
        # NOTE(dtantsur): this is called before logging is initialized, so we
        # cannot really log a warning here.
        pass
else:
    _translators = i18n.TranslatorFactory(domain='ironic')

    # The primary translation function using the well-known name "_"
    _ = _translators.primary

    install = i18n.install



# Copyright 2013,2014 Cray Inc
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Ironic SNMP power manager.

Provides basic power control using an SNMP-enabled smart power controller.
Uses a pluggable driver model to support devices with different SNMP object
models.

"""

import abc
import time

from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import importutils

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import states
from ironic.common import utils
from ironic.conductor import task_manager
from ironic.conf import CONF
from ironic.drivers import base

pysnmp = importutils.try_import('pysnmp')
if pysnmp:
    from pysnmp import error as snmp_error
    from pysnmp import hlapi as snmp

    snmp_auth_protocols = {
        'md5': snmp.usmHMACMD5AuthProtocol,
        'sha': snmp.usmHMACSHAAuthProtocol,
        'sha224': snmp.usmHMAC128SHA224AuthProtocol,
        'sha256': snmp.usmHMAC192SHA256AuthProtocol,
        'sha384': snmp.usmHMAC256SHA384AuthProtocol,
        'sha512': snmp.usmHMAC384SHA512AuthProtocol,
        'none': snmp.usmNoAuthProtocol,
    }

    snmp_priv_protocols = {
        'des': snmp.usmDESPrivProtocol,
        '3des': snmp.usm3DESEDEPrivProtocol,
        'aes': snmp.usmAesCfb128Protocol,
        'aes192': snmp.usmAesCfb192Protocol,
        'aes256': snmp.usmAesCfb256Protocol,
        'aes192blmt': snmp.usmAesBlumenthalCfb192Protocol,
        'aes256blmt': snmp.usmAesBlumenthalCfb256Protocol,
        'none': snmp.usmNoPrivProtocol,
    }

else:
    snmp = None
    snmp_error = None

    snmp_auth_protocols = {
        'none': None
    }

    snmp_priv_protocols = {
        'none': None
    }

LOG = logging.getLogger(__name__)


SNMP_V1 = '1'
SNMP_V2C = '2c'
SNMP_V3 = '3'
SNMP_PORT = 161

REQUIRED_PROPERTIES = {
    'snmp_driver': _("PDU manufacturer driver.  Required."),
    'snmp_address': _("PDU IPv4 address or hostname.  Required."),
    'snmp_outlet': _("PDU power outlet index (1-based).  Required."),
}
OPTIONAL_PROPERTIES = {
    'snmp_version':
        _("SNMP protocol version: %(v1)s, %(v2c)s or %(v3)s  "
          "(optional, default %(v1)s).")
        % {"v1": SNMP_V1, "v2c": SNMP_V2C, "v3": SNMP_V3},
    'snmp_port':
        _("SNMP port, default %(port)d.") % {"port": SNMP_PORT},
    'snmp_community':
        _("SNMP community name to use for read and/or write class SNMP "
          "commands unless `snmp_community_read` and/or "
          "`snmp_community_write` properties are present in which case the "
          "latter takes over. Applicable only to versions %(v1)s and %(v2c)s.")
        % {"v1": SNMP_V1, "v2c": SNMP_V2C},
    'snmp_community_read':
        _("SNMP community name to use for read class SNMP commands. "
          "Takes precedence over the `snmp_community` property. "
          "Applicable only to versions %(v1)s and %(v2c)s.")
        % {"v1": SNMP_V1, "v2c": SNMP_V2C},
    'snmp_community_write':
        _("SNMP community name to use for write class SNMP commands. "
          "Takes precedence over the `snmp_community` property. "
          "Applicable only to versions %(v1)s and %(v2c)s.")
        % {"v1": SNMP_V1, "v2c": SNMP_V2C},
    'snmp_user':
        _("SNMPv3 User-based Security Model (USM) username. "
          "Required for version %(v3)s.")
        % {"v3": SNMP_V3},
    'snmp_auth_protocol':
        _("SNMPv3 message authentication protocol ID. "
          "Known values are: %(auth)s. "
          "Default is 'none' unless 'snmp_auth_key' is provided. "
          "In the latter case 'md5' is the default.")
        % {'auth': sorted(snmp_auth_protocols)},
    'snmp_auth_key':
        _("SNMPv3 message authentication key. "
          "Must be 8+ characters long. "
          "Required when message authentication is used. "
          "This key is used by the 'snmp_auth_protocol' algorithm."),
    'snmp_priv_protocol':
        _("SNMPv3 message privacy (encryption) protocol ID. "
          "Known values are: %(priv)s. "
          "Using message privacy requires using message authentication. "
          "Default is 'none' unless 'snmp_priv_key' is provided. "
          "In the latter case 'des' is the default.")
        % {'priv': sorted(snmp_priv_protocols)},
    'snmp_priv_key':
        _("SNMPv3 message authentication key. "
          "Must be 8+ characters long. "
          "Required when message authentication is used. "
          "This key is used by the 'snmp_priv_protocol' algorithm."),
    'snmp_context_engine_id':
        _("SNMPv3 context engine ID. "
          "Default is the value of authoritative engine ID."),
    'snmp_context_name':
        _("SNMPv3 context name. "
          "Default is an empty string ('')."),
}

DEPRECATED_PROPERTIES = {
    # synonym for `snmp_user`
    'snmp_security':
        _("SNMPv3 User-based Security Model (USM) username. "
          "Required for version %(v3)s. "
          "This property is deprecated, please use `snmp_user` instead.")
        % {"v3": SNMP_V3},
}

COMMON_PROPERTIES = REQUIRED_PROPERTIES.copy()
COMMON_PROPERTIES.update(OPTIONAL_PROPERTIES)
COMMON_PROPERTIES.update(DEPRECATED_PROPERTIES)


class SNMPClient(object):
    """SNMP client object.

    Performs low level SNMP get and set operations. Encapsulates all
    interaction with PySNMP to simplify dynamic importing and unit testing.
    """

    def __init__(self, address, port, version,
                 read_community=None, write_community=None,
                 user=None, auth_proto=None,
                 auth_key=None, priv_proto=None,
                 priv_key=None, context_engine_id=None, context_name=None):
        if not snmp:
            raise exception.DriverLoadError(
                driver=self.__class__.__name__,
                reason=_("Unable to import python-pysnmp library")
            )

        self.address = address
        self.port = port
        self.version = version
        if self.version == SNMP_V3:
            self.user = user
            self.auth_proto = auth_proto
            self.auth_key = auth_key
            self.priv_proto = priv_proto
            self.priv_key = priv_key
        else:
            self.read_community = read_community
            self.write_community = write_community

        self.context_engine_id = context_engine_id
        self.context_name = context_name or ''

        self.snmp_engine = snmp.SnmpEngine()

    def _get_auth(self, write_mode=False):
        """Return the authorization data for an SNMP request.

        :param write_mode: `True` if write class SNMP command is
            executed. Default is `False`.
        :returns: Either
            :class:`pysnmp.hlapi.CommunityData`
            or :class:`pysnmp.hlapi.UsmUserData`
            object depending on SNMP version being used.
        """
        if self.version == SNMP_V3:
            # NOTE(TheJulia): Ignore Bandit error B509 argument parsing as
            # the check is for a count of <3 arguments, however our line
            # wrapping causes the check to trigger.
            return snmp.UsmUserData(  # nosec B509
                self.user,
                authKey=self.auth_key,
                authProtocol=self.auth_proto,
                privKey=self.priv_key,
                privProtocol=self.priv_proto
            )

        else:
            mp_model = 1 if self.version == SNMP_V2C else 0
            return snmp.CommunityData(
                self.write_community if write_mode else self.read_community,
                mpModel=mp_model
            )

    def _get_transport(self):
        """Return the transport target for an SNMP request.

        :returns: A :class:
            `pysnmp.hlapi.UdpTransportTarget` object.
        :raises: :class:`pysnmp.error.PySnmpError` if the transport address
            is bad.
        """
        # The transport target accepts timeout and retries parameters, which
        # default to 1 (second) and 5 respectively. These are deemed sensible
        # enough to allow for an unreliable network or slow device.
        return snmp.UdpTransportTarget(
            (self.address, self.port),
            timeout=CONF.snmp.udp_transport_timeout,
            retries=CONF.snmp.udp_transport_retries)

    def _get_context(self):
        """Return the SNMP context for an SNMP request.

        :returns: A :class:
            `pysnmp.hlapi.ContextData` object.
        :raises: :class:`pysnmp.error.PySnmpError` if SNMP context data
            is bad.
        """
        return snmp.ContextData(
            contextEngineId=self.context_engine_id,
            contextName=self.context_name
        )

    def get(self, oid):
        """Use PySNMP to perform an SNMP GET operation on a single object.

        :param oid: The OID of the object to get.
        :raises: SNMPFailure if an SNMP request fails.
        :returns: The value of the requested object.
        """
        try:
            snmp_gen = snmp.getCmd(self.snmp_engine,
                                   self._get_auth(),
                                   self._get_transport(),
                                   self._get_context(),
                                   snmp.ObjectType(snmp.ObjectIdentity(oid)))

        except snmp_error.PySnmpError as e:
            raise exception.SNMPFailure(operation="GET", error=e)

        error_indication, error_status, error_index, var_binds = next(snmp_gen)

        if error_indication:
            # SNMP engine-level error.
            raise exception.SNMPFailure(operation="GET",
                                        error=error_indication)

        if error_status:
            # SNMP PDU error.
            raise exception.SNMPFailure(operation="GET",
                                        error=error_status.prettyPrint())

        # We only expect a single value back
        name, val = var_binds[0]
        return val

    def get_next(self, oid):
        """Use PySNMP to perform an SNMP GET NEXT operation on a table object.

        :param oid: The OID of the object to get.
        :raises: SNMPFailure if an SNMP request fails.
        :returns: A list of values of the requested table object.
        """
        try:
            snmp_gen = snmp.nextCmd(self.snmp_engine,
                                    self._get_auth(),
                                    self._get_transport(),
                                    self._get_context(),
                                    snmp.ObjectType(snmp.ObjectIdentity(oid)),
                                    lexicographicMode=False)

        except snmp_error.PySnmpError as e:
            raise exception.SNMPFailure(operation="GET_NEXT", error=e)

        vals = []
        for (error_indication, error_status, error_index,
                var_binds) in snmp_gen:

            if error_indication:
                # SNMP engine-level error.
                raise exception.SNMPFailure(operation="GET_NEXT",
                                            error=error_indication)

            if error_status:
                # SNMP PDU error.
                raise exception.SNMPFailure(operation="GET_NEXT",
                                            error=error_status.prettyPrint())

            # this is not a table, but a table row
            # e.g. 1-D array of tuples
            _name, value = var_binds[0]
            vals.append(value)

        return vals

    def set(self, oid, value):
        """Use PySNMP to perform an SNMP SET operation on a single object.

        :param oid: The OID of the object to set.
        :param value: The value of the object to set.
        :raises: SNMPFailure if an SNMP request fails.
        """
        try:
            snmp_gen = snmp.setCmd(self.snmp_engine,
                                   self._get_auth(write_mode=True),
                                   self._get_transport(),
                                   self._get_context(),
                                   snmp.ObjectType(
                                       snmp.ObjectIdentity(oid), value))

        except snmp_error.PySnmpError as e:
            raise exception.SNMPFailure(operation="SET", error=e)

        error_indication, error_status, error_index, var_binds = next(snmp_gen)

        if error_indication:
            # SNMP engine-level error.
            raise exception.SNMPFailure(operation="SET",
                                        error=error_indication)

        if error_status:
            # SNMP PDU error.
            raise exception.SNMPFailure(operation="SET",
                                        error=error_status.prettyPrint())


def _get_client(snmp_info):
    """Create and return an SNMP client object.

    :param snmp_info: SNMP driver info.
    :returns: A :class:`SNMPClient` object.
    """
    return SNMPClient(snmp_info["address"],
                      snmp_info["port"],
                      snmp_info["version"],
                      snmp_info.get("read_community"),
                      snmp_info.get("write_community"),
                      snmp_info.get("user"),
                      snmp_info.get("auth_protocol"),
                      snmp_info.get("auth_key"),
                      snmp_info.get("priv_protocol"),
                      snmp_info.get("priv_key"),
                      snmp_info.get("context_engine_id"),
                      snmp_info.get("context_name"))


_memoized = {}


def memoize(f):
    def memoized(self, node_info):
        hashable_node_info = frozenset((key, val)
                                       for key, val in node_info.items()
                                       if key != 'outlet')
        if hashable_node_info not in _memoized:
            _memoized[hashable_node_info] = f(self)
        return _memoized[hashable_node_info]
    return memoized


def retry_on_outdated_cache(f):
    def wrapper(self):
        try:
            return f(self)

        except exception.SNMPFailure:
            hashable_node_info = (
                frozenset((key, val)
                          for key, val in self.snmp_info.items()
                          if key != 'outlet')
            )
            del _memoized[hashable_node_info]
            self.driver = self._get_pdu_driver(self.snmp_info)
            return f(self)

    return wrapper


class SNMPDriverBase(object, metaclass=abc.ABCMeta):
    """SNMP power driver base class.

    The SNMPDriver class hierarchy implements manufacturer-specific MIB actions
    over SNMP to interface with different smart power controller products.
    """

    oid_enterprise = (1, 3, 6, 1, 4, 1)
    retry_interval = 1

    def __init__(self, snmp_info):
        self.snmp_info = snmp_info
        self.client = _get_client(snmp_info)

    @abc.abstractmethod
    def _snmp_power_state(self):
        """Perform the SNMP request required to get the current power state.

        :raises: SNMPFailure if an SNMP request fails.
        :returns: power state. One of :class:`ironic.common.states`.
        """

    @abc.abstractmethod
    def _snmp_power_on(self):
        """Perform the SNMP request required to set the power on.

        :raises: SNMPFailure if an SNMP request fails.
        """

    @abc.abstractmethod
    def _snmp_power_off(self):
        """Perform the SNMP request required to set the power off.

        :raises: SNMPFailure if an SNMP request fails.
        """

    def _snmp_wait_for_state(self, goal_state):
        """Wait for the power state of the PDU outlet to change.

        :param goal_state: The power state to wait for, one of
            :class:`ironic.common.states`.
        :raises: SNMPFailure if an SNMP request fails.
        :returns: power state. One of :class:`ironic.common.states`.
        """

        def _poll_for_state(mutable):
            """Called at an interval until the node's power is consistent.

            :param mutable: dict object containing "state" and "next_time"
            :raises: SNMPFailure if an SNMP request fails.
            """
            mutable["state"] = self._snmp_power_state()
            if mutable["state"] == goal_state:
                raise loopingcall.LoopingCallDone()

            mutable["next_time"] += self.retry_interval
            if mutable["next_time"] >= CONF.snmp.power_timeout:
                mutable["state"] = states.ERROR
                raise loopingcall.LoopingCallDone()

        # Pass state to the looped function call in a mutable form.
        state = {"state": None, "next_time": 0}
        timer = loopingcall.FixedIntervalLoopingCall(_poll_for_state,
                                                     state)
        timer.start(interval=self.retry_interval).wait()
        LOG.debug("power state '%s'", state["state"])
        return state["state"]

    def power_state(self):
        """Returns a node's current power state.

        :raises: SNMPFailure if an SNMP request fails.
        :returns: power state. One of :class:`ironic.common.states`.
        """
        return self._snmp_power_state()

    def power_on(self):
        """Set the power state to this node to ON.

        :raises: SNMPFailure if an SNMP request fails.
        :returns: power state. One of :class:`ironic.common.states`.
        """
        time.sleep(CONF.snmp.power_action_delay)
        self._snmp_power_on()
        return self._snmp_wait_for_state(states.POWER_ON)

    def power_off(self):
        """Set the power state to this node to OFF.

        :raises: SNMPFailure if an SNMP request fails.
        :returns: power state. One of :class:`ironic.common.states`.
        """
        self._snmp_power_off()
        time.sleep(CONF.snmp.power_action_delay)
        return self._snmp_wait_for_state(states.POWER_OFF)

    def power_reset(self):
        """Reset the power to this node.

        :raises: SNMPFailure if an SNMP request fails.
        :returns: power state. One of :class:`ironic.common.states`.
        """
        power_result = self.power_off()
        if power_result != states.POWER_OFF:
            return states.ERROR
        time.sleep(CONF.snmp.reboot_delay)
        power_result = self.power_on()
        if power_result != states.POWER_ON:
            return states.ERROR
        return power_result


class SNMPDriverSimple(SNMPDriverBase):
    """SNMP driver base class for simple PDU devices.

    Here, simple refers to devices which provide a single SNMP object for
    controlling the power state of an outlet.

    The default OID of the power state object is of the form
    <enterprise OID>.<device OID>.<outlet ID>. A different OID may be specified
    by overriding the _snmp_oid method in a subclass.
    """

    def __init__(self, *args, **kwargs):
        super(SNMPDriverSimple, self).__init__(*args, **kwargs)
        self.oid = self._snmp_oid()

    @property
    @abc.abstractmethod
    def oid_device(self):
        """Device dependent portion of the power state object OID."""

    @property
    @abc.abstractmethod
    def value_power_on(self):
        """Value representing power on state."""

    @property
    @abc.abstractmethod
    def value_power_off(self):
        """Value representing power off state."""

    def _snmp_oid(self):
        """Return the OID of the power state object.

        :returns: Power state object OID as a tuple of integers.
        """
        outlet = self.snmp_info['outlet']
        return self.oid_enterprise + self.oid_device + (outlet,)

    def _snmp_power_state(self):
        state = self.client.get(self.oid)

        # Translate the state to an Ironic power state.
        if state == self.value_power_on:
            power_state = states.POWER_ON
        elif state == self.value_power_off:
            power_state = states.POWER_OFF
        else:
            LOG.warning("SNMP PDU %(addr)s outlet %(outlet)s: "
                        "unrecognised power state %(state)s.",
                        {'addr': self.snmp_info['address'],
                         'outlet': self.snmp_info['outlet'],
                         'state': state})
            power_state = states.ERROR

        return power_state

    def _snmp_power_on(self):
        value = snmp.Integer(self.value_power_on)
        self.client.set(self.oid, value)

    def _snmp_power_off(self):
        value = snmp.Integer(self.value_power_off)
        self.client.set(self.oid, value)


class SNMPDriverAten(SNMPDriverSimple):
    """SNMP driver class for Aten PDU devices.

    SNMP objects for Aten PDU:
    1.3.6.1.4.1.21317.1.3.2.2.2.2 Outlet Power
    Values: 1=Off, 2=On, 3=Pending, 4=Reset
    """
    system_id = (21317,)
    oid_device = (21317, 1, 3, 2, 2, 2, 2)
    value_power_on = 2
    value_power_off = 1

    def _snmp_oid(self):
        """Return the OID of the power state object.

        :returns: Power state object OID as a tuple of integers.
        """
        outlet = self.snmp_info['outlet']
        return self.oid_enterprise + self.oid_device + (outlet, 0,)


class SNMPDriverAPCMasterSwitch(SNMPDriverSimple):
    """SNMP driver class for APC MasterSwitch PDU devices.

    SNMP objects for APC SNMPDriverAPCMasterSwitch PDU:
    1.3.6.1.4.1.318.1.1.4.4.2.1.3 sPDUOutletCtl
    Values: 1=On, 2=Off, 3=PowerCycle, [...more options follow]
    """

    system_id = (318, 1, 1, 4)
    oid_device = (318, 1, 1, 4, 4, 2, 1, 3)
    value_power_on = 1
    value_power_off = 2


class SNMPDriverAPCMasterSwitchPlus(SNMPDriverSimple):
    """SNMP driver class for APC MasterSwitchPlus PDU devices.

    SNMP objects for APC SNMPDriverAPCMasterSwitchPlus PDU:
    1.3.6.1.4.1.318.1.1.6.5.1.1.5 sPDUOutletControlMSPOutletCommand
    Values: 1=On, 3=Off, [...more options follow]
    """

    system_id = (318, 1, 1, 6)
    oid_device = (318, 1, 1, 6, 5, 1, 1, 5)
    value_power_on = 1
    value_power_off = 3


class SNMPDriverAPCRackPDU(SNMPDriverSimple):
    """SNMP driver class for APC RackPDU devices.

    SNMP objects for APC SNMPDriverAPCRackPDU PDU:
    # 1.3.6.1.4.1.318.1.1.12.3.3.1.1.4 rPDUOutletControlOutletCommand
    Values: 1=On, 2=Off, 3=PowerCycle, [...more options follow]
    """

    system_id = (318, 1, 1, 12)
    oid_device = (318, 1, 1, 12, 3, 3, 1, 1, 4)
    value_power_on = 1
    value_power_off = 2


class SNMPDriverCyberPower(SNMPDriverSimple):
    """SNMP driver class for CyberPower PDU devices.

    SNMP objects for CyberPower PDU:
    1.3.6.1.4.1.3808.1.1.3.3.3.1.1.4 ePDUOutletControlOutletCommand
    Values: 1=On, 2=Off, 3=PowerCycle, [...more options follow]
    """

    # NOTE(mgoddard): This device driver is currently untested, this driver has
    #                 been implemented based upon its published MIB
    #                 documentation.

    system_id = (3808,)
    oid_device = (3808, 1, 1, 3, 3, 3, 1, 1, 4)
    value_power_on = 1
    value_power_off = 2


class SNMPDriverTeltronix(SNMPDriverSimple):
    """SNMP driver class for Teltronix PDU devices.

    SNMP objects for Teltronix PDU:
    1.3.6.1.4.1.23620.1.2.2.1.4   Outlet Power
    Values: 1=Off, 2=On
    """

    system_id = (23620,)
    oid_device = (23620, 1, 2, 2, 1, 4)
    value_power_on = 2
    value_power_off = 1


class SNMPDriverEatonPower(SNMPDriverBase):
    """SNMP driver class for Eaton Power PDU.

    The Eaton power PDU does not follow the model of SNMPDriverSimple as it
    uses multiple SNMP objects.

    SNMP objects for Eaton Power PDU
    1.3.6.1.4.1.534.6.6.7.6.6.1.2.<outlet ID> outletControlStatus
    Read 0=off, 1=on, 2=pending off, 3=pending on
    1.3.6.1.4.1.534.6.6.7.6.6.1.3.<outlet ID> outletControlOffCmd
    Write 0 for immediate power off
    1.3.6.1.4.1.534.6.6.7.6.6.1.4.<outlet ID> outletControlOnCmd
    Write 0 for immediate power on
    """

    # NOTE(mgoddard): This device driver is currently untested, this driver has
    #                 been implemented based upon its published MIB
    #                 documentation.

    system_id = (534,)
    oid_device = (534, 6, 6, 7, 6, 6, 1)
    oid_status = (2,)
    oid_poweron = (3,)
    oid_poweroff = (4,)

    status_off = 0
    status_on = 1
    status_pending_off = 2
    status_pending_on = 3

    value_power_on = 0
    value_power_off = 0

    def __init__(self, *args, **kwargs):
        super(SNMPDriverEatonPower, self).__init__(*args, **kwargs)
        # Due to its use of different OIDs for different actions, we only form
        # an OID that holds the common substring of the OIDs for power
        # operations.
        self.oid_base = self.oid_enterprise + self.oid_device

    def _snmp_oid(self, oid):
        """Return the OID for one of the outlet control objects.

        :param oid: The action-dependent portion of the OID, as a tuple of
            integers.
        :returns: The full OID as a tuple of integers.
        """
        outlet = self.snmp_info['outlet']
        return self.oid_base + oid + (outlet,)

    def _snmp_power_state(self):
        oid = self._snmp_oid(self.oid_status)
        state = self.client.get(oid)

        # Translate the state to an Ironic power state.
        if state in (self.status_on, self.status_pending_off):
            power_state = states.POWER_ON
        elif state in (self.status_off, self.status_pending_on):
            power_state = states.POWER_OFF
        else:
            LOG.warning("Eaton Power SNMP PDU %(addr)s outlet %(outlet)s: "
                        "unrecognised power state %(state)s.",
                        {'addr': self.snmp_info['address'],
                         'outlet': self.snmp_info['outlet'],
                         'state': state})
            power_state = states.ERROR

        return power_state

    def _snmp_power_on(self):
        oid = self._snmp_oid(self.oid_poweron)
        value = snmp.Integer(self.value_power_on)
        self.client.set(oid, value)

    def _snmp_power_off(self):
        oid = self._snmp_oid(self.oid_poweroff)
        value = snmp.Integer(self.value_power_off)
        self.client.set(oid, value)


class SNMPDriverBaytechMRP27(SNMPDriverSimple):
    """SNMP driver class for Baytech MRP27 PDU devices.

    SNMP objects for Baytech MRP27 PDU:
    4779, 1, 3, 5, 3, 1, 3, {unit_id}  Outlet Power
    Values: 0=Off, 1=On, 2=Reboot
    """

    # TODO(srobert): Add support for dynamically allocated unit_id when needed
    unit_id = 1
    oid_device = (4779, 1, 3, 5, 3, 1, 3) + (unit_id,)
    value_power_off = 0
    value_power_on = 1


class SNMPDriverServerTechSentry3(SNMPDriverBase):
    """SNMP driver class for Server Technology Sentry 3 PDUs.

    ftp://ftp.servertech.com/Pub/SNMP/sentry3/Sentry3.mib

    SNMP objects for Server Technology Power PDU.
    1.3.6.1.4.1.1718.3.2.3.1.5.1.1.<outlet ID> outletStatus
    Read 0=off, 1=on, 2=off wait, 3=on wait, [...more options follow]
    1.3.6.1.4.1.1718.3.2.3.1.11.1.1.<outlet ID> outletControlAction
    Write 0=no action, 1=on, 2=off, 3=reboot
    """

    oid_device = (1718, 3, 2, 3, 1)
    oid_tower_infeed_idx = (1, 1, )
    oid_power_status = (5,)
    oid_power_action = (11,)

    status_off = 0
    status_on = 1
    status_off_wait = 2
    status_on_wait = 3

    value_power_on = 1
    value_power_off = 2

    def __init__(self, *args, **kwargs):
        super(SNMPDriverServerTechSentry3, self).__init__(*args, **kwargs)
        # Due to its use of different OIDs for different actions, we only form
        # an OID that holds the common substring of the OIDs for power
        # operations.
        self.oid_base = self.oid_enterprise + self.oid_device

    def _snmp_oid(self, oid):
        """Return the OID for one of the outlet control objects.

        :param oid: The action-dependent portion of the OID, as a tuple of
            integers.
        :returns: The full OID as a tuple of integers.
        """

        outlet = self.snmp_info['outlet']
        full_oid = self.oid_base + oid + self.oid_tower_infeed_idx + (outlet,)
        return full_oid

    def _snmp_power_state(self):
        oid = self._snmp_oid(self.oid_power_status)
        state = self.client.get(oid)

        # Translate the state to an Ironic power state.
        if state in (self.status_on, self.status_off_wait):
            power_state = states.POWER_ON
        elif state in (self.status_off, self.status_on_wait):
            power_state = states.POWER_OFF
        else:
            LOG.warning("SeverTech Sentry3 PDU %(addr)s oid %(oid) outlet "
                        "%(outlet)s: unrecognised power state %(state)s.",
                        {'addr': self.snmp_info['address'],
                         'oid': oid,
                         'outlet': self.snmp_info['outlet'],
                         'state': state})
            power_state = states.ERROR

        return power_state

    def _snmp_power_on(self):
        oid = self._snmp_oid(self.oid_power_action)
        value = snmp.Integer(self.value_power_on)
        self.client.set(oid, value)

    def _snmp_power_off(self):
        oid = self._snmp_oid(self.oid_power_action)
        value = snmp.Integer(self.value_power_off)
        self.client.set(oid, value)


class SNMPDriverServerTechSentry4(SNMPDriverBase):
    """SNMP driver class for Server Technology Sentry 4 PDUs.

    https://www.servertech.com/support/sentry-mib-oid-tree-downloads

    SNMP objects for Server Technology Power PDU.
    1.3.6.1.4.1.1718.4.1.8.5.1.1<outlet ID> outletStatus
    notSet (0) fixedOn (1) idleOff (2) idleOn (3) [...more options follow]
    pendOn (8) pendOff (9) off (10) on (11) [...more options follow]
    eventOff (16) eventOn (17) eventReboot (18) eventShutdown (19)
    1.3.6.1.4.1.1718.4.1.8.5.1.2.<outlet ID> outletControlAction
    Write 0=no action, 1=on, 2=off, 3=reboot
    """

    oid_device = (1718, 4, 1, 8, 5, 1)
    oid_tower_infeed_idx = (1, 1, )
    oid_power_status = (1,)
    oid_power_action = (2,)

    notSet = 0
    fixedOn = 1
    idleOff = 2
    idleOn = 3
    wakeOff = 4
    wakeOn = 5
    ocpOff = 6
    ocpOn = 7
    status_pendOn = 8
    status_pendOff = 9
    status_off = 10
    status_on = 11
    reboot = 12
    shutdown = 13
    lockedOff = 14
    lockedOn = 15

    value_power_on = 1
    value_power_off = 2

    def __init__(self, *args, **kwargs):
        super(SNMPDriverServerTechSentry4, self).__init__(*args, **kwargs)
        # Due to its use of different OIDs for different actions, we only form
        # an OID that holds the common substring of the OIDs for power
        # operations.
        self.oid_base = self.oid_enterprise + self.oid_device

    def _snmp_oid(self, oid):
        """Return the OID for one of the outlet control objects.

        :param oid: The action-dependent portion of the OID, as a tuple of
            integers.
        :returns: The full OID as a tuple of integers.
        """

        outlet = self.snmp_info['outlet']
        full_oid = self.oid_base + oid + self.oid_tower_infeed_idx + (outlet,)
        return full_oid

    def _snmp_power_state(self):
        oid = self._snmp_oid(self.oid_power_status)
        state = self.client.get(oid)

        # Translate the state to an Ironic power state.
        if state in (self.status_on, self.status_pendOn, self.idleOn):
            power_state = states.POWER_ON
        elif state in (self.status_off, self.status_pendOff):
            power_state = states.POWER_OFF
        else:
            LOG.warning("ServerTech Sentry4 PDU %(addr)s oid %(oid)s outlet "
                        "%(outlet)s: unrecognised power state %(state)s.",
                        {'addr': self.snmp_info['address'],
                         'oid': oid,
                         'outlet': self.snmp_info['outlet'],
                         'state': state})
            power_state = states.ERROR

        return power_state

    def _snmp_power_on(self):
        oid = self._snmp_oid(self.oid_power_action)
        value = snmp.Integer(self.value_power_on)
        self.client.set(oid, value)

    def _snmp_power_off(self):
        oid = self._snmp_oid(self.oid_power_action)
        value = snmp.Integer(self.value_power_off)
        self.client.set(oid, value)


class SNMPDriverRaritanPDU2(SNMPDriverBase):
    """SNMP driver class for Raritan PDU2 PDUs.

    http://support.raritan.com/px2/version-2.4.1/mibs/pdu2-mib-020400-39592.txt
    http://cdn.raritan.com/download/PX/v1.5.20/PDU-MIB.txt

    Command:
        snmpset -v2c -c private -m+PDU2-MIB <pdu IP address> \
            PDU2-MIB::switchingOperation.1.4 = cycle
        snmpset -v2c -c private <pdu IP address> \
            .1.3.6.1.4.1.13742.6.4.1.2.1.2.1.4 i 2
    Output:
        PDU2-MIB::switchingOperation.1.4 = INTEGER: cycle(2)
    """

    oid_device = (13742, 6, 4, 1, 2, 1)
    oid_power_action = (2, )
    oid_power_status = (3, )
    oid_tower_infeed_idx = (1, )

    unavailable = -1
    status_open = 0
    status_closed = 1
    belowLowerCritical = 2
    belowLowerWarning = 3
    status_normal = 4
    aboveUpperWarning = 5
    aboveUpperCritical = 6
    status_on = 7
    status_off = 8
    detected = 9
    notDetected = 10
    alarmed = 11
    ok = 12
    marginal = 13
    fail = 14
    yes = 15
    no = 16
    standby = 17
    one = 18
    two = 19
    inSync = 20
    outOfSync = 21

    value_power_on = 1
    value_power_off = 0

    def __init__(self, *args, **kwargs):
        super(SNMPDriverRaritanPDU2, self).__init__(*args, **kwargs)
        # Due to its use of different OIDs for different actions, we only form
        # an OID that holds the common substring of the OIDs for power
        # operations.
        self.oid_base = self.oid_enterprise + self.oid_device

    def _snmp_oid(self, oid):
        """Return the OID for one of the outlet control objects.

        :param oid: The action-dependent portion of the OID, as a tuple of
            integers.
        :returns: The full OID as a tuple of integers.
        """

        outlet = self.snmp_info['outlet']
        full_oid = self.oid_base + oid + self.oid_tower_infeed_idx + (outlet,)
        return full_oid

    def _snmp_power_state(self):
        oid = self._snmp_oid(self.oid_power_status)
        state = self.client.get(oid)

        # Translate the state to an Ironic power state.
        if state == self.status_on:
            power_state = states.POWER_ON
        elif state == self.status_off:
            power_state = states.POWER_OFF
        else:
            LOG.warning("Raritan PDU2 PDU %(addr)s oid %(oid)s outlet "
                        "%(outlet)s: unrecognised power state %(state)s.",
                        {'addr': self.snmp_info['address'],
                         'oid': oid,
                         'outlet': self.snmp_info['outlet'],
                         'state': state})
            power_state = states.ERROR

        return power_state

    def _snmp_power_on(self):
        oid = self._snmp_oid(self.oid_power_action)
        value = snmp.Integer(self.value_power_on)
        self.client.set(oid, value)

    def _snmp_power_off(self):
        oid = self._snmp_oid(self.oid_power_action)
        value = snmp.Integer(self.value_power_off)
        self.client.set(oid, value)


class SNMPDriverVertivGeistPDU(SNMPDriverBase):
    """SNMP driver class for VertivGeist NU30017L/NU30019L PDU.

    https://mibs.observium.org/mib/GEIST-V5-MIB/

    """

    oid_device = (21239, 5, 2, 3, 5, 1)
    oid_power_action = (6, )
    oid_power_status = (4, )
    oid_tower_infeed_idx = (1, )

    on = 1
    off = 2
    on2off = 3
    off2on = 4
    rebootOn = 5
    rebootOff = 5
    unavailable = 7

    value_power_on = 2
    value_power_off = 4

    def __init__(self, *args, **kwargs):
        super(SNMPDriverVertivGeistPDU, self).__init__(*args, **kwargs)
        # Due to its use of different OIDs for different actions, we only form
        # an OID that holds the common substring of the OIDs for power
        # operations.
        self.oid_base = self.oid_enterprise + self.oid_device

    def _snmp_oid(self, oid):
        """Return the OID for one of the outlet control objects.

        :param oid: The action-dependent portion of the OID, as a tuple of
            integers.

        :returns: The full OID as a tuple of integers.
        """

        outlet = self.snmp_info['outlet']
        full_oid = self.oid_base + oid + (outlet,)
        return full_oid

    def _snmp_power_state(self):
        oid = self._snmp_oid(self.oid_power_status)
        state = self.client.get(oid)

        # Translate the state to an Ironic power state.
        if state in (self.on, self.on2off):
            power_state = states.POWER_ON
        elif state in (self.off, self.off2on):
            power_state = states.POWER_OFF
        else:
            LOG.warning("Vertiv Geist PDU %(addr)s oid %(oid)s outlet "
                        "%(outlet)s: unrecognised power state %(state)s.",
                        {'addr': self.snmp_info['address'],
                         'oid': oid,
                         'outlet': self.snmp_info['outlet'],
                         'state': state})
            power_state = states.ERROR

        return power_state

    def _snmp_power_on(self):
        oid = self._snmp_oid(self.oid_power_action)
        value = snmp.Integer(self.value_power_on)
        self.client.set(oid, value)

    def _snmp_power_off(self):
        oid = self._snmp_oid(self.oid_power_action)
        value = snmp.Integer(self.value_power_off)
        self.client.set(oid, value)


class SNMPDriverAuto(SNMPDriverBase):

    SYS_OBJ_OID = (1, 3, 6, 1, 2, 1, 1, 2)

    def __init__(self, *args, **kwargs):
        super(SNMPDriverAuto, self).__init__(*args, **kwargs)
        self.driver = self._get_pdu_driver(*args, **kwargs)

    def _get_pdu_driver(self, *args, **kwargs):
        drivers_map = {}

        for name, obj in DRIVER_CLASSES.items():
            if not getattr(obj, 'system_id', False):
                continue

            system_id = self.oid_enterprise + getattr(obj, 'system_id')

            if (system_id in drivers_map
                    and drivers_map[system_id] is not obj):
                raise exception.InvalidParameterValue(_(
                    "SNMPDriverAuto: duplicate driver system ID prefix "
                    "%(system_id)s") % {'system_id': system_id})

            drivers_map[system_id] = obj
            LOG.debug("SNMP driver mapping %(system_id)s -> %(name)s",
                      {'system_id': system_id, 'name': obj.__name__})

        system_id = self._fetch_driver(*args, **kwargs)

        LOG.debug("SNMP device reports sysObjectID %(system_id)s",
                  {'system_id': system_id})

        system_id_prefix = tuple(system_id)

        # pick driver by the longest matching sysObjectID prefix
        while len(system_id_prefix) > len(self.oid_enterprise):
            try:
                Driver = drivers_map[system_id_prefix]
                LOG.debug("Chosen SNMP driver %(name)s based on sysObjectID "
                          "prefix %(system_id_prefix)s", {Driver.__name__,
                                                          system_id_prefix})
                return Driver(*args, **kwargs)

            except KeyError:
                system_id_prefix = system_id_prefix[:-1]

        raise exception.InvalidParameterValue(_(
            "SNMPDriverAuto: no driver matching %(system_id)s") %
            {'system_id': system_id})

    @retry_on_outdated_cache
    def _snmp_power_state(self):
        current_power_state = self.driver._snmp_power_state()
        return current_power_state

    @retry_on_outdated_cache
    def _snmp_power_on(self):
        return self.driver._snmp_power_on()

    @retry_on_outdated_cache
    def _snmp_power_off(self):
        return self.driver._snmp_power_off()

    @memoize
    def _fetch_driver(self):
        return self.client.get(self.SYS_OBJ_OID)


# A dictionary of supported drivers keyed by snmp_driver attribute
DRIVER_CLASSES = {
    'apc': SNMPDriverAPCMasterSwitch,
    'apc_masterswitch': SNMPDriverAPCMasterSwitch,
    'apc_masterswitchplus': SNMPDriverAPCMasterSwitchPlus,
    'apc_rackpdu': SNMPDriverAPCRackPDU,
    'aten': SNMPDriverAten,
    'cyberpower': SNMPDriverCyberPower,
    'eatonpower': SNMPDriverEatonPower,
    'teltronix': SNMPDriverTeltronix,
    'baytech_mrp27': SNMPDriverBaytechMRP27,
    'servertech_sentry3': SNMPDriverServerTechSentry3,
    'servertech_sentry4': SNMPDriverServerTechSentry4,
    'raritan_pdu2': SNMPDriverRaritanPDU2,
    'vertivgeist_pdu': SNMPDriverVertivGeistPDU,
    'auto': SNMPDriverAuto,
}


def _parse_driver_info_snmpv3_user(node, info):
    snmp_info = {}

    if 'snmp_user' not in info and 'snmp_security' not in info:
        raise exception.MissingParameterValue(_(
            "SNMP driver requires `driver_info/snmp_user` to be set in "
            "node %(node)s configuration for SNMP version %(ver)s.") %
            {'node': node.uuid, 'ver': SNMP_V3})

    snmp_info['user'] = info.get('snmp_user', info.get('snmp_security'))

    if 'snmp_security' in info:
        LOG.warning("The `driver_info/snmp_security` parameter is deprecated "
                    "in favor of `driver_info/snmp_user` parameter. Please "
                    "remove the `driver_info/snmp_security` parameter from "
                    "node %(node)s configuration.", {'node': node.uuid})

        if 'snmp_user' in info:
            LOG.warning("The `driver_info/snmp_security` parameter is ignored "
                        "in favor of `driver_info/snmp_user` parameter in "
                        "node %(node)s configuration.", {'node': node.uuid})

    return snmp_info


def _parse_driver_info_snmpv3_crypto(node, info):
    snmp_info = {}

    if 'snmp_auth_protocol' in info:
        auth_p = info['snmp_auth_protocol']
        try:
            snmp_info['auth_protocol'] = snmp_auth_protocols[auth_p]

        except KeyError:
            raise exception.InvalidParameterValue(_(
                "SNMPPowerDriver: unknown SNMPv3 authentication protocol "
                "`driver_info/snmp_auth_protocol` %(proto)s in node %(node)s "
                "configuration, known protocols are: %(protos)s") %
                {'node': node.uuid, 'proto': auth_p,
                 'protos': ', '.join(snmp_auth_protocols)}
            )
    if 'snmp_priv_protocol' in info:
        priv_p = info['snmp_priv_protocol']
        try:
            snmp_info['priv_protocol'] = snmp_priv_protocols[priv_p]

        except KeyError:
            raise exception.InvalidParameterValue(_(
                "SNMPPowerDriver: unknown SNMPv3 privacy protocol "
                "`driver_info/snmp_priv_protocol` %(proto)s in node "
                "%(node)s configuration, known protocols are: %(protos)s") %
                {'node': node.uuid, 'proto': priv_p,
                 'protos': ', '.join(snmp_priv_protocols)}
            )
    if 'snmp_auth_key' in info:
        auth_k = info['snmp_auth_key']
        if len(auth_k) < 8:
            raise exception.InvalidParameterValue(_(
                "SNMPPowerDriver: short SNMPv3 authentication key "
                "`driver_info/snmp_auth_key` in node %(node)s configuration "
                "(8+ chars required)") % {'node': node.uuid})

        snmp_info['auth_key'] = auth_k

        if 'auth_protocol' not in snmp_info:
            snmp_info['auth_protocol'] = snmp_auth_protocols['md5']

    if 'snmp_priv_key' in info:
        priv_k = info['snmp_priv_key']
        if len(priv_k) < 8:
            raise exception.InvalidParameterValue(_(
                "SNMPPowerDriver: short SNMPv3 privacy key "
                "`driver_info/snmp_priv_key` node %(node)s configuration "
                "(8+ chars required)") % {'node': node.uuid})

        snmp_info['priv_key'] = priv_k

        if 'priv_protocol' not in snmp_info:
            snmp_info['priv_protocol'] = snmp_priv_protocols['des']

    if ('priv_protocol' in snmp_info
            and 'auth_protocol' not in snmp_info):
        raise exception.MissingParameterValue(_(
            "SNMPPowerDriver: SNMPv3 privacy requires authentication. "
            "Please add `driver_info/auth_protocol` property to node "
            "%(node)s configuration.") % {'node': node.uuid})

    if ('auth_protocol' in snmp_info
            and 'auth_key' not in snmp_info):
        raise exception.MissingParameterValue(_(
            "SNMPPowerDriver: missing SNMPv3 authentication key while "
            "`driver_info/snmp_auth_protocol` is present. Please "
            "add `driver_info/snmp_auth_key` to node %(node)s "
            "configuration.") % {'node': node.uuid})

    if ('priv_protocol' in snmp_info
            and 'priv_key' not in snmp_info):
        raise exception.MissingParameterValue(_(
            "SNMPPowerDriver: missing SNMPv3 privacy key while "
            "`driver_info/snmp_priv_protocol` is present. Please "
            "add `driver_info/snmp_priv_key` to node %(node)s "
            "configuration.") % {'node': node.uuid})

    return snmp_info


def _parse_driver_info_snmpv3_context(node, info):
    snmp_info = {}

    if 'snmp_context_engine_id' in info:
        snmp_info['context_engine_id'] = info['snmp_context_engine_id']

    if 'snmp_context_name' in info:
        snmp_info['context_name'] = info['snmp_context_name']

    return snmp_info


def _parse_driver_info(node):
    """Parse a node's driver_info values.

    Return a dictionary of validated driver information, usable for
    SNMPDriver object creation.

    :param node: An Ironic node object.
    :returns: SNMP driver info.
    :raises: MissingParameterValue if any required parameters are missing.
    :raises: InvalidParameterValue if any parameters are invalid.
    """
    info = node.driver_info or {}
    missing_info = [key for key in REQUIRED_PROPERTIES if not info.get(key)]
    if missing_info:
        raise exception.MissingParameterValue(_(
            "SNMP driver requires the following parameters to be set in "
            "node's driver_info: %s.") % missing_info)

    snmp_info = {}

    # Validate PDU driver type
    snmp_info['driver'] = info['snmp_driver']
    if snmp_info['driver'] not in DRIVER_CLASSES:
        raise exception.InvalidParameterValue(_(
            "SNMPPowerDriver: unknown driver: '%s'") % snmp_info['driver'])

    # In absence of a version, default to SNMPv1
    snmp_info['version'] = info.get('snmp_version', SNMP_V1)
    if snmp_info['version'] not in (SNMP_V1, SNMP_V2C, SNMP_V3):
        raise exception.InvalidParameterValue(_(
            "SNMPPowerDriver: unknown SNMP version: '%s'") %
            snmp_info['version'])

    # In absence of a configured UDP port, default to the standard port
    port_str = info.get('snmp_port', SNMP_PORT)
    snmp_info['port'] = utils.validate_network_port(port_str, 'snmp_port')

    if snmp_info['port'] < 1 or snmp_info['port'] > 65535:
        raise exception.InvalidParameterValue(_(
            "SNMPPowerDriver: SNMP UDP port out of range: %d")
            % snmp_info['port'])

    # Extract version-dependent required parameters
    if snmp_info['version'] in (SNMP_V1, SNMP_V2C):
        read_community = info.get('snmp_community_read')
        if read_community is None:
            read_community = info.get('snmp_community')

        write_community = info.get('snmp_community_write')
        if write_community is None:
            write_community = info.get('snmp_community')

        if not read_community or not write_community:
            raise exception.MissingParameterValue(_(
                "SNMP driver requires `snmp_community` or "
                "`snmp_community_read`/`snmp_community_write` properties "
                "to be set for version %s.") % snmp_info['version'])
        snmp_info['read_community'] = read_community
        snmp_info['write_community'] = write_community

    elif snmp_info['version'] == SNMP_V3:
        snmp_info.update(_parse_driver_info_snmpv3_user(node, info))
        snmp_info.update(_parse_driver_info_snmpv3_crypto(node, info))
        snmp_info.update(_parse_driver_info_snmpv3_context(node, info))

    # Target PDU IP address and power outlet identification
    snmp_info['address'] = info['snmp_address']
    outlet = info['snmp_outlet']
    try:
        snmp_info['outlet'] = int(outlet)
    except ValueError:
        raise exception.InvalidParameterValue(_(
            "SNMPPowerDriver: PDU power outlet index is not an integer: %s")
            % outlet)

    return snmp_info


def _get_driver(node):
    """Return a new SNMP driver object of the correct type for `node`.

    :param node: Single node object.
    :raises: InvalidParameterValue if node power config is incomplete or
        invalid.
    :returns: SNMP driver object.
    """
    snmp_info = _parse_driver_info(node)
    cls = DRIVER_CLASSES[snmp_info['driver']]
    return cls(snmp_info)


class SNMPPower(base.PowerInterface):
    """SNMP Power Interface.

    This PowerInterface class provides a mechanism for controlling the power
    state of a physical device using an SNMP-enabled smart power controller.
    """

    def get_properties(self):
        """Return the properties of the interface.

        :returns: dictionary of <property name>:<property description> entries.
        """
        return COMMON_PROPERTIES

    def validate(self, task):
        """Check that node.driver_info contains the requisite fields.

        :raises: MissingParameterValue if required SNMP parameters are missing.
        :raises: InvalidParameterValue if SNMP parameters are invalid.
        """
        _parse_driver_info(task.node)

    def get_power_state(self, task):
        """Get the current power state.

        Poll the SNMP device for the current power state of the node.

        :param task: An instance of `ironic.manager.task_manager.TaskManager`.
        :raises: MissingParameterValue if required SNMP parameters are missing.
        :raises: InvalidParameterValue if SNMP parameters are invalid.
        :raises: SNMPFailure if an SNMP request fails.
        :returns: power state. One of :class:`ironic.common.states`.
        """
        driver = _get_driver(task.node)
        power_state = driver.power_state()
        return power_state

    @task_manager.require_exclusive_lock
    def set_power_state(self, task, pstate, timeout=None):
        """Turn the power on or off.

        Set the power state of a node.

        :param task: An instance of `ironic.manager.task_manager.TaskManager`.
        :param pstate: Either POWER_ON or POWER_OFF from :class:
            `ironic.common.states`.
        :param timeout: timeout (in seconds). Unsupported by this interface.
        :raises: MissingParameterValue if required SNMP parameters are missing.
        :raises: InvalidParameterValue if SNMP parameters are invalid or
            `pstate` is invalid.
        :raises: PowerStateFailure if the final power state of the node is not
            as requested after the timeout.
        :raises: SNMPFailure if an SNMP request fails.
        """
        # TODO(rloo): Support timeouts!
        if timeout is not None:
            LOG.warning(
                "The 'snmp' Power Interface's 'set_power_state' method "
                "doesn't support the 'timeout' parameter. Ignoring "
                "timeout=%(timeout)s",
                {'timeout': timeout})

        driver = _get_driver(task.node)
        if pstate == states.POWER_ON:
            state = driver.power_on()
        elif pstate == states.POWER_OFF:
            state = driver.power_off()
        else:
            raise exception.InvalidParameterValue(_("set_power_state called "
                                                    "with invalid power "
                                                    "state %s.") % str(pstate))
        if state != pstate:
            raise exception.PowerStateFailure(pstate=pstate)

    @task_manager.require_exclusive_lock
    def reboot(self, task, timeout=None):
        """Cycles the power to a node.

        :param task: An instance of `ironic.manager.task_manager.TaskManager`.
        :param timeout: timeout (in seconds). Unsupported by this interface.
        :raises: MissingParameterValue if required SNMP parameters are missing.
        :raises: InvalidParameterValue if SNMP parameters are invalid.
        :raises: PowerStateFailure if the final power state of the node is not
            POWER_ON after the timeout.
        :raises: SNMPFailure if an SNMP request fails.
        """
        # TODO(rloo): Support timeouts!
        if timeout is not None:
            LOG.warning("The 'snmp' Power Interface's 'reboot' method "
                        "doesn't support the 'timeout' parameter. Ignoring "
                        "timeout=%(timeout)s",
                        {'timeout': timeout})

        driver = _get_driver(task.node)
        state = driver.power_reset()
        if state != states.POWER_ON:
            raise exception.PowerStateFailure(pstate=states.POWER_ON)
